/*
 * wx_cmds.cc
 * wxWindows command processing
 *
 */

#include "wx.h"
#include "wx_hash.h"
#include "wx_types.h"

#include <iostream.h>
#include <ctype.h>
#include <stdlib.h>

#include "clipscpp.h"
#include "wx_cmds.h"

#include "clips.h"

ClipsErrorClass ClipsError;
wxHashTable wxClipsObjects(wxKEY_INTEGER);

// Record what frames and dialogs are used so we can clear
// them up if necessary
wxList wxClipsFrameList;

/*
 * Mechanism for getting CLIPS callbacks to return a value
 *
 */

int clipsReturnType = clipsUNKNOWN;
char *clipsReturnString = NULL;
long clipsReturnLong = 0;
double clipsReturnDouble = 0.0;
extern "C" int clipsEchoOn;

/*
 * The argument to an externally called function is passed to this
 * function, which stores the appropriate return value and type
 * for the external program to fetch.
 */

void clipsReturnResult()
{
  DATA_OBJECT temp;
  RtnUnknown(1, &temp);

  switch (GetType(temp))
  {
    case INTEGER:
    {
      clipsReturnLong = DOToLong(temp);
      clipsReturnType = clipsLONG;
      break;
    }
    case STRING:
    case SYMBOL:
    {
      if (clipsReturnString)
        delete clipsReturnString;

      clipsReturnString = copystring(DOToString(temp));
      clipsReturnType = clipsSTRING;
      break;
    }
    case FLOAT:
    {
      clipsReturnDouble = DOToDouble(temp);
      clipsReturnType = clipsDOUBLE;
      break;
    }
    default:
    {
      clipsReturnType = clipsUNKNOWN;
    }
  }
}

/*
 * Get platform:
 *  "Windows 3.1", "XView" or "Motif"
 *
 */

char *clipsGetPlatform()
{
#ifdef wx_motif
  return (char *)AddSymbol("Motif");
#endif
#ifdef wx_xview
  return (char *)AddSymbol("XView");
#endif
#ifdef wx_msw
  return (char *)AddSymbol("Windows 3.1");
#endif
}

/************************************************************************
 * CLEARING-UP FUNCTIONS
 *
 ************************************************************************/

void wxCleanWindows(void)
{
  wxNode *node = wxClipsFrameList.First();
  while (node)
  {
    wxWindow *win = (wxWindow *)node->Data();
    win->Show(FALSE);
    delete win;
    node = wxClipsFrameList.First();
  }
}

void clipsCleanWindows()
{
  if (ArgCountCheck("clipsCleanWindows", EXACTLY, 0) == -1)
    return;

  wxCleanWindows();
}


/************************************************************************
 * SPECIAL-PURPOSE CLIPS WINDOWS/WIDGETS
 *
 ************************************************************************/

class wxClipsCallbacks: public wxList
{
 public:
  wxClipsCallbacks(void):wxList(wxKEY_STRING)
    {}
  ~wxClipsCallbacks(void)
   {
     wxNode *node = First();
     while (node)
     {
       delete (char *)node->Data();
       node = node->Next();
     }
   }
};


// Dialog box
class wxClipsDialogBox: public wxDialogBox
{
 public:
  wxClipsDialogBox(wxFrame *frame, char *title,
              Bool modal = FALSE,
              int x = -1, int y = -1,
              int width = -1, int height = -1):
   wxDialogBox(frame, title, modal, x, y, width, height)
   {
     wxClipsFrameList.Append(this);
   }
                   
  ~wxClipsDialogBox()
    {
      wxClipsCallbacks *data;
      if (data = (wxClipsCallbacks *)GetClientData())
        delete data;
      wxDeleteThingInTable((long)this);
      wxClipsFrameList.DeleteObject(this);
    }

  void OnSize(int w, int h)
  {
    wxClipsCallbacks *callbacks = (wxClipsCallbacks *)GetClientData();
    if (callbacks)
    {
      wxNode *node = callbacks->Find("OnSize");
      if (node)
      {
        char *func = (char *)node->Data();
        char buf[100];
        sprintf(buf, "(%s %ld %d %d)", func, (long)this, w, h);
        clipsEchoOn = FALSE;
        RouteCommand(buf);
        clipsEchoOn = TRUE;
      }
    }
    else
      wxDialogBox::OnSize(w, h);
  }
};

// Panel
class wxClipsPanel: public wxPanel
{
 public:
  wxClipsPanel(wxFrame *frame,
              int x = -1, int y = -1,
              int width = -1, int height = -1, int style = 0):
   wxPanel(frame, x, y, width, height, style)
    { }
                   
  ~wxClipsPanel()
    {
      wxClipsCallbacks *data;
      if (data = (wxClipsCallbacks *)GetClientData())
        delete data;
      wxDeleteThingInTable((long)this);
    }

  void OnSize(int w, int h)
  {
    wxClipsCallbacks *callbacks = (wxClipsCallbacks *)GetClientData();
    if (callbacks)
    {
      wxNode *node = callbacks->Find("OnSize");
      if (node)
      {
        char *func = (char *)node->Data();
        char buf[100];
        sprintf(buf, "(%s %ld %d %d)", func, (long)this, w, h);
        clipsEchoOn = FALSE;
        RouteCommand(buf);
        clipsEchoOn = TRUE;
      }
    }
    else
      wxPanel::OnSize(w, h);
  }
};

// Canvas
class wxClipsCanvas: public wxCanvas
{
 public:
  wxClipsCanvas(wxFrame *frame,
              int x = -1, int y = -1,
              int width = -1, int height = -1, int style = 0):
   wxCanvas(frame, x, y, width, height, style)
   {
     wxCanvasDC *dc = GetDC();
     wxStoreThingInTable((long)dc, dc, wxTYPE_DC_CANVAS);
   }
                   
  ~wxClipsCanvas()
    {
      wxClipsCallbacks *data;
      if (data = (wxClipsCallbacks *)GetClientData())
        delete data;
      wxDeleteThingInTable((long)this);
      wxCanvasDC *dc = GetDC();
      wxDeleteThingInTable((long)dc);
    }

  void OnSize(int w, int h)
  {
    wxClipsCallbacks *callbacks = (wxClipsCallbacks *)GetClientData();
    if (callbacks)
    {
      wxNode *node = callbacks->Find("OnSize");
      if (node)
      {
        char *func = (char *)node->Data();
        char buf[100];
        sprintf(buf, "(%s %ld %d %d)", func, (long)this, w, h);
        clipsEchoOn = FALSE;
        RouteCommand(buf);
        clipsEchoOn = TRUE;
      }
    }
//    wxCanvas::OnSize(w, h);
  }

  void OnChar(wxKeyEvent& event)
  {
    wxClipsCallbacks *callbacks = (wxClipsCallbacks *)GetClientData();
    if (callbacks)
    {
      wxNode *node = callbacks->Find("OnChar");
      if (node)
      {
        wxStoreThingInTable((long)&event, &event, wxTYPE_KEY_EVENT);

        char *func = (char *)node->Data();
        char buf[100];
        sprintf(buf, "(%s %ld %ld %ld)", func, (long)this, event.KeyCode(), (long)&event);
        clipsEchoOn = FALSE;
        RouteCommand(buf);
        clipsEchoOn = TRUE;

        wxDeleteThingInTable((long)&event);
        return;
      }
    }
    wxCanvas::OnChar(event);
  }

  void OnEvent(wxMouseEvent& event)
  {
    wxClipsCallbacks *callbacks = (wxClipsCallbacks *)GetClientData();
    if (callbacks)
    {
      wxNode *node = callbacks->Find("OnEvent");
      if (node)
      {
        wxStoreThingInTable((long)&event, &event, wxTYPE_MOUSE_EVENT);

        char *func = (char *)node->Data();
        char buf[100];
        sprintf(buf, "(%s %ld %ld)", func, (long)this, (long)&event);
        clipsEchoOn = FALSE;
        RouteCommand(buf);
        clipsEchoOn = TRUE;
        
        wxDeleteThingInTable((long)&event);
        return;
      }
    }
//    wxCanvas::OnEvent(event);
  }

  void OnPaint(void)
  {
    wxClipsCallbacks *callbacks = (wxClipsCallbacks *)GetClientData();
    if (callbacks)
    {
      wxNode *node = callbacks->Find("OnPaint");
      if (node)
      {
        char *func = (char *)node->Data();
        char buf[100];
        sprintf(buf, "(%s %ld)", func, (long)this);
        clipsEchoOn = FALSE;
        RouteCommand(buf);
        clipsEchoOn = TRUE;
        return;
      }
    }
//    wxCanvas::OnPaint();
  }
};

// Pen
class wxClipsPen: public wxPen
{
 public:
  wxClipsPen(wxColour& col, int width, int style):
   wxPen(col, width, style)
   {
   }
                   
  ~wxClipsPen()
   {
      wxDeleteThingInTable((long)this);
   }
};

// Brush
class wxClipsBrush: public wxBrush
{
 public:
  wxClipsBrush(wxColour& col, int style):
   wxBrush(col, style)
   {
   }
                   
  ~wxClipsBrush()
   {
      wxDeleteThingInTable((long)this);
   }
};

// Font
class wxClipsFont: public wxFont
{
 public:
  wxClipsFont(int pointSize, int family, int size, int height):
   wxFont(pointSize, family, size, height)
   {
   }
                   
  ~wxClipsFont()
   {
      wxDeleteThingInTable((long)this);
   }
};

// Icon
class wxClipsIcon: public wxIcon
{
 public:
  wxClipsIcon(char *file):
   wxIcon(file)
   {
   }
                   
  ~wxClipsIcon()
   {
      wxDeleteThingInTable((long)this);
   }
};

    
// Text subwindow
class wxClipsTextWindow: public wxTextWindow
{
 public:
  wxClipsTextWindow(wxFrame *frame,
              int x = -1, int y = -1,
              int width = -1, int height = -1, int style = 0):
   wxTextWindow(frame, x, y, width, height, style)
    { }
                   
  ~wxClipsTextWindow()
    {
      wxClipsCallbacks *data;
      if (data = (wxClipsCallbacks *)GetClientData())
        delete data;
      wxDeleteThingInTable((long)this);
    }

  void OnSize(int w, int h)
  {
    wxClipsCallbacks *callbacks = (wxClipsCallbacks *)GetClientData();
    if (callbacks)
    {
      wxNode *node = callbacks->Find("OnSize");
      if (node)
      {
        char *func = (char *)node->Data();
        char buf[100];
        sprintf(buf, "(%s %ld %d %d)", func, (long)this, w, h);
        clipsEchoOn = FALSE;
        RouteCommand(buf);
        clipsEchoOn = TRUE;
      }
    }
    else
      wxTextWindow::OnSize(w, h);
  }
};

// Frame
class wxClipsFrame: public wxFrame
{
 public:
  wxClipsFrame(wxFrame *parent, char *title,
          int x=-1, int y=-1, int width=-1, int height=-1,
          int style = wxSDI):wxFrame(parent, title, x, y, width, height, style)
  {
    wxClipsFrameList.Append(this);
  }

  ~wxClipsFrame()
    {
      wxClipsCallbacks *data;
      if (data = (wxClipsCallbacks *)GetClientData())
        delete data;
      wxDeleteThingInTable((long)this);
      wxClipsFrameList.DeleteObject(this);
    }

  void OnSize(int w, int h)
  {
    wxClipsCallbacks *callbacks = (wxClipsCallbacks *)GetClientData();
    if (callbacks)
    {
      wxNode *node = callbacks->Find("OnSize");
      if (node)
      {
        char *func = (char *)node->Data();
        char buf[100];
        sprintf(buf, "(%s %ld %d %d)", func, (long)this, w, h);
        clipsEchoOn = FALSE;
        RouteCommand(buf);
        clipsEchoOn = TRUE;
      }
    }
    else
      wxFrame::OnSize(w, h);
  }

  void OnMenuCommand(int id)
  {
    wxClipsCallbacks *callbacks = (wxClipsCallbacks *)GetClientData();
    if (callbacks)
    {
      wxNode *node = callbacks->Find("OnMenuCommand");
      if (node)
      {
        char *func = (char *)node->Data();
        char buf[100];
        sprintf(buf, "(%s %ld %d)", func, (long)this, (long)id);
        clipsEchoOn = FALSE;
        RouteCommand(buf);
        clipsEchoOn = TRUE;
      }
    }
    else
      wxFrame::OnMenuCommand(id);
  }

  Bool OnClose(void)
  {
    wxClipsCallbacks *callbacks = (wxClipsCallbacks *)GetClientData();
    if (callbacks)
    {
      wxNode *node = callbacks->Find("OnClose");
      if (node)
      {
        char *func = (char *)node->Data();
        char buf[200];
        clipsReturnType = clipsUNKNOWN;
        sprintf(buf, "(return-result (%s %ld))", func, (long)this);
        clipsEchoOn = FALSE;
        RouteCommand(buf);
        clipsEchoOn = TRUE;
        
        if (clipsReturnType != clipsLONG)
	{
          ClipsError << "OnClose callback for frame should return an integer (1 or 0)!\n";
          return FALSE;
	}
        else return (Bool)clipsReturnLong;
      }
    }
    return wxFrame::OnClose();
  }
};

// Button
class wxClipsButton: public wxButton
{
 public:
  char *clipsFunc;
  wxClipsButton(wxPanel *panel, char *clipsFuncName, wxFunction func, char *label, int x = -1, int y = -1,
           int width = -1, int height = -1):wxButton(panel, func, label, x, y, width, height)
 {
    if (clipsFuncName)
      clipsFunc = copystring(clipsFuncName);
    else clipsFunc = NULL;
 }
 ~wxClipsButton(void)
 {
   if (clipsFunc)
     delete clipsFunc;
   wxDeleteThingInTable((long)this);
 }
};

// Message
class wxClipsMessage: public wxMessage
{
 public:
  wxClipsMessage(wxPanel *panel, char *label, int x = -1, int y = -1):
    wxMessage(panel, label, x, y)
 {
 }
 ~wxClipsMessage(void)
 {
   wxDeleteThingInTable((long)this);
 }
};

// Text
class wxClipsText: public wxText
{
 public:
  char *clipsFunc;
  wxClipsText(wxPanel *panel, char *clipsFuncName, wxFunction func, char *label, char *value = "",
         int x = -1, int y = -1, int width = -1, int height = -1):
    wxText(panel, func, label, value, x, y, width, height)
 {
    if (clipsFuncName)
      clipsFunc = copystring(clipsFuncName);
    else clipsFunc = NULL;
 }

 ~wxClipsText(void)
 {
   if (clipsFunc)
     delete clipsFunc;
   wxDeleteThingInTable((long)this);
 }
};

// Multi-line Text
class wxClipsMultiText: public wxMultiText
{
 public:
  char *clipsFunc;
  wxClipsMultiText(wxPanel *panel, char *clipsFuncName, wxFunction func, char *label, char *value = "",
         int x = -1, int y = -1, int width = -1, int height = -1):
    wxMultiText(panel, func, label, value, x, y, width, height)
 {
    if (clipsFuncName)
      clipsFunc = copystring(clipsFuncName);
    else clipsFunc = NULL;
 }

 ~wxClipsMultiText(void)
 {
   if (clipsFunc)
     delete clipsFunc;
   wxDeleteThingInTable((long)this);
 }
};

// Check box
class wxClipsCheckBox: public wxCheckBox
{
 public:
  char *clipsFunc;
  wxClipsCheckBox(wxPanel *panel, char *clipsFuncName, wxFunction func, char *label,
         int x = -1, int y = -1, int width = -1, int height = -1):
    wxCheckBox(panel, func, label, x, y, width, height)
 {
    if (clipsFuncName)
      clipsFunc = copystring(clipsFuncName);
    else clipsFunc = NULL;
 }

 ~wxClipsCheckBox(void)
 {
   if (clipsFunc)
     delete clipsFunc;
   wxDeleteThingInTable((long)this);
 }
};

// Slider
class wxClipsSlider: public wxSlider
{
 public:
  char *clipsFunc;
  wxClipsSlider(wxPanel *panel, char *clipsFuncName, wxFunction func, char *label,
         int value, int min_value, int max_value, int width, int x = -1, int y = -1):
    wxSlider(panel, func, label, value, min_value, max_value, width, x, y)
 {
    if (clipsFuncName)
      clipsFunc = copystring(clipsFuncName);
    else clipsFunc = NULL;
 }

 ~wxClipsSlider(void)
 {
   if (clipsFunc)
     delete clipsFunc;
   wxDeleteThingInTable((long)this);
 }
};

// List box
class wxClipsListBox: public wxListBox
{
 public:
  char *clipsFunc;
  wxClipsListBox(wxPanel *panel, char *clipsFuncName, wxFunction func, char *label,
                 Bool multiple = FALSE, int x = -1, int y = -1, int width = -1, int height = -1):
    wxListBox(panel, func, label, multiple, x, y, width, height)
 {
    if (clipsFuncName)
      clipsFunc = copystring(clipsFuncName);
    else clipsFunc = NULL;
 }

 ~wxClipsListBox(void)
 {
   if (clipsFunc)
     delete clipsFunc;
   wxDeleteThingInTable((long)this);
 }
};

// Choice
class wxClipsChoice: public wxChoice
{
 public:
  char *clipsFunc;
  wxClipsChoice(wxPanel *panel, char *clipsFuncName, wxFunction func, char *label,
                 int x = -1, int y = -1, int width = -1, int height = -1):
    wxChoice(panel, func, label, x, y, width, height)
 {
    if (clipsFuncName)
      clipsFunc = copystring(clipsFuncName);
    else clipsFunc = NULL;
 }

 ~wxClipsChoice(void)
 {
   if (clipsFunc)
     delete clipsFunc;
   wxDeleteThingInTable((long)this);
 }
};

class wxClipsMenuBar: public wxMenuBar
{
 public:
  wxClipsMenuBar(void)
    { }
  ~wxClipsMenuBar(void)
  {
     wxDeleteThingInTable((long)this);
  }
};

class wxClipsMenu: public wxMenu
{
 public:
  wxClipsMenu(char *title):wxMenu(title)
    { }
  ~wxClipsMenu(void)
  {
     wxDeleteThingInTable((long)this);
  }
};


/************************************************************************
 * GENERIC WINDOW FUNCTIONS
 *
 ************************************************************************/


// Delete window
long clipsWindowDelete()
{
  if (ArgCountCheck("clipsWindowDelete", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    delete win;
    wxDeleteThingInTable(id);
    return 1;
  }
  else
  {
    ClipsError << "Error in window-delete: could not find window " << id << ".\n";
    return 0;
  }
}

// Show window
long clipsWindowShow()
{
  if (ArgCountCheck("clipsWindowShow", AT_LEAST, 1) == -1)
    return 0;

  long id = RtnLong(1);
  Bool show = TRUE;
  if (RtnArgCount() > 1)
    show = (Bool)RtnLong(2);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    win->Show(show);
    return 1;
  }
  else
  {
    ClipsError << "Error in window-show: could not find window " << id << ".\n";
    return 0;
  }
}

// Get window parent
long clipsWindowGetParent()
{
  if (ArgCountCheck("clipsWindowGetParent", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    wxWindow *parent = win->GetParent();
    if (parent)
      return (long)parent;
    else
      return 0;
  }
  else
  {
    ClipsError << "Error in window-get-parent: could not find window " << id << ".\n";
    return 0;
  }
}

// Add a callback to a window
long clipsWindowAddCallback()
{
  if (ArgCountCheck("clipsWindowAddCallback", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);

  char *event_type = RtnLexeme(2);
  char *callback = RtnLexeme(3);

  if (!event_type || !callback)
    return 0;

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    wxClipsCallbacks *data = (wxClipsCallbacks *)win->GetClientData();
    if (!data)
    {
      data = new wxClipsCallbacks;
      win->SetClientData((char *)data);
    }
    data->Append(copystring(event_type), (wxObject *)copystring(callback));
    
    return 1;
  }
  else
  {
    ClipsError << "Error in window-add-callback: could not find window " << id << ".\n";
    return 0;
  }
}

// Get window width
long clipsWindowGetWidth()
{
  if (ArgCountCheck("clipsWindowGetWidth", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    int w, h;
    win->GetSize(&w, &h);
    return w;
  }
  else
  {
    ClipsError << "Error in window-get-width: could not find window " << id << ".\n";
    return 0;
  }
}

// Get window height
long clipsWindowGetHeight()
{
  if (ArgCountCheck("clipsWindowGetHeight", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    int w, h;
    win->GetSize(&w, &h);
    return h;
  }
  else
  {
    ClipsError << "Error in window-get-height: could not find window " << id << ".\n";
    return 0;
  }
}

// Get window client width
long clipsWindowGetClientWidth()
{
  if (ArgCountCheck("clipsWindowGetClientWidth", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    int w, h;
    win->GetClientSize(&w, &h);
    return w;
  }
  else
  {
    ClipsError << "Error in window-get-client-width: could not find window " << id << ".\n";
    return 0;
  }
}

// Get window client height
long clipsWindowGetClientHeight()
{
  if (ArgCountCheck("clipsWindowGetClientHeight", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    int w, h;
    win->GetClientSize(&w, &h);
    return h;
  }
  else
  {
    ClipsError << "Error in window-get-client-height: could not find window " << id << ".\n";
    return 0;
  }
}

// Get window x position
long clipsWindowGetX()
{
  if (ArgCountCheck("clipsWindowGetX", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    int x, y;
    win->GetPosition(&x, &y);
    return x;
  }
  else
  {
    ClipsError << "Error in window-get-x: could not find window " << id << ".\n";
    return 0;
  }
}

// Get window y position
long clipsWindowGetY()
{
  if (ArgCountCheck("clipsWindowGetY", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    int x, y;
    win->GetPosition(&x, &y);
    return y;
  }
  else
  {
    ClipsError << "Error in window-get-y: could not find window " << id << ".\n";
    return 0;
  }
}

// Set size
long clipsWindowSetSize()
{
  if (ArgCountCheck("clipsWindowSetSize", EXACTLY, 5) == -1)
    return 0;

  long id = RtnLong(1);
  int x = (int) RtnLong(2);
  int y = (int) RtnLong(3);
  int w = (int) RtnLong(4);
  int h = (int) RtnLong(5);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    win->SetSize(x, y, w, h);
    return 1;
  }
  else
  {
    ClipsError << "Error in window-set-size: could not find window " << id << ".\n";
    return 0;
  }
}

// Set client size
long clipsWindowSetClientSize()
{
  if (ArgCountCheck("clipsWindowSetClientSize", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  int w = (int) RtnLong(2);
  int h = (int) RtnLong(3);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    win->SetClientSize(w, h);
    return 1;
  }
  else
  {
    ClipsError << "Error in window-set-client-size: could not find window " << id << ".\n";
    return 0;
  }
}

// Set focus
long clipsWindowSetFocus()
{
  if (ArgCountCheck("clipsWindowSetFocus", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    win->SetFocus();
    return 1;
  }
  else
  {
    ClipsError << "Error in window-set-focus: could not find window " << id << ".\n";
    return 0;
  }
}

// Fit
long clipsWindowFit()
{
  if (ArgCountCheck("clipsWindowFit", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    win->Fit();
    return 1;
  }
  else
  {
    ClipsError << "Error in window-fit: could not find window " << id << ".\n";
    return 0;
  }
}

// Centre
long clipsWindowCentre()
{
  if (ArgCountCheck("clipsWindowCentre", AT_LEAST, 1) == -1)
    return 0;

  int no_args = RtnArgCount();
  long id = RtnLong(1);
  int direction = wxHORIZONTAL;
  if (no_args > 1)
    direction = (int) RtnLong(2);

  wxWindow *win = (wxWindow *)wxGetTypedObject(id, wxTYPE_WINDOW);
  if (win)
  {
    win->Centre(direction);
    return 1;
  }
  else
  {
    ClipsError << "Error in window-centre: could not find window " << id << ".\n";
    return 0;
  }
}

/************************************************************************
 * GENERIC PANEL ITEM FUNCTIONS
 *
 ************************************************************************/

// Set this button to be the default
long clipsPanelItemSetDefault()
{
  if (ArgCountCheck("clipsPanelItemSetDefault", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxButton *item = (wxButton *)wxGetTypedObject(id, wxTYPE_BUTTON);
  if (item)
  {
    item->SetDefault();
    return 1;
  }
  else
  {
    ClipsError << "Error in panel-item-set-default: could not find item " << id << ".\n";
    return 0;
  }
}

// Set label
long clipsPanelItemSetLabel()
{
  if (ArgCountCheck("clipsPanelItemSetLabel", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *label = RtnLexeme(2);
  if (!label)
    return 0;

  wxItem *item = (wxItem *)wxGetTypedObject(id, wxTYPE_ITEM);
  if (item)
  {
    item->SetLabel(label);
    return 1;
  }
  else
  {
    ClipsError << "Error in panel-item-set-label: could not find item " << id << ".\n";
    return 0;
  }
}

// Get label
char *clipsPanelItemGetLabel()
{
  if (ArgCountCheck("clipsPanelItemSetLabel", EXACTLY, 1) == -1)
    return (char *)AddSymbol("");

  long id = RtnLong(1);

  wxItem *item = (wxItem *)wxGetTypedObject(id, wxTYPE_ITEM);
  if (item)
  {
    char *label = item->GetLabel();
    return (char *)AddSymbol(label);
  }
  else
  {
    ClipsError << "Error in panel-item-get-label: could not find item " << id << ".\n";
    return (char *)AddSymbol("");
  }
}

/************************************************************************
 * PANEL ITEM CALLBACKS
 *
 ************************************************************************/

// Event within scope of a callback (if nested, can only rely on
// getting event info immediately after function is called.)
wxCommandEvent *wxCurrentEvent = NULL;

void wxClipsButtonCallback(wxClipsButton& item, wxCommandEvent& event)
{
  wxCurrentEvent = &event;
  if (item.clipsFunc)
  {
    char buf[100];
    sprintf(buf, "(%s %ld)", item.clipsFunc, &item);
    clipsEchoOn = FALSE;
    RouteCommand(buf);
    clipsEchoOn = TRUE;
  }
  wxCurrentEvent = NULL;
}

void wxClipsTextCallback(wxClipsText& item, wxCommandEvent& event)
{
  wxCurrentEvent = &event;
  if (item.clipsFunc)
  {
    char buf[100];
    sprintf(buf, "(%s %ld)", item.clipsFunc, &item);
    clipsEchoOn = FALSE;
    RouteCommand(buf);
    clipsEchoOn = TRUE;
  }
  wxCurrentEvent = NULL;
}

void wxClipsMultiTextCallback(wxClipsMultiText& item, wxCommandEvent& event)
{
  wxCurrentEvent = &event;
  if (item.clipsFunc)
  {
    char buf[100];
    sprintf(buf, "(%s %ld)", item.clipsFunc, &item);
    clipsEchoOn = FALSE;
    RouteCommand(buf);
    clipsEchoOn = TRUE;
  }
  wxCurrentEvent = NULL;
}

void wxClipsListBoxCallback(wxClipsListBox& item, wxCommandEvent& event)
{
  wxCurrentEvent = &event;
  if (item.clipsFunc)
  {
    char buf[100];
    sprintf(buf, "(%s %ld)", item.clipsFunc, &item);
    clipsEchoOn = FALSE;
    RouteCommand(buf);
    clipsEchoOn = TRUE;
  }
  wxCurrentEvent = NULL;
}

void wxClipsChoiceCallback(wxClipsChoice& item, wxCommandEvent& event)
{
  wxCurrentEvent = &event;
  if (item.clipsFunc)
  {
    char buf[100];
    sprintf(buf, "(%s %ld)", item.clipsFunc, &item);
    clipsEchoOn = FALSE;
    RouteCommand(buf);
    clipsEchoOn = TRUE;
  }
  wxCurrentEvent = NULL;
}

void wxClipsCheckBoxCallback(wxClipsCheckBox& item, wxCommandEvent& event)
{
  wxCurrentEvent = &event;
  if (item.clipsFunc)
  {
    char buf[100];
    sprintf(buf, "(%s %ld)", item.clipsFunc, &item);
    clipsEchoOn = FALSE;
    RouteCommand(buf);
    clipsEchoOn = TRUE;
  }
  wxCurrentEvent = NULL;
}

void wxClipsSliderCallback(wxClipsSlider& item, wxCommandEvent& event)
{
  wxCurrentEvent = &event;
  if (item.clipsFunc)
  {
    char buf[100];
    sprintf(buf, "(%s %ld)", item.clipsFunc, &item);
    clipsEchoOn = FALSE;
    RouteCommand(buf);
    clipsEchoOn = TRUE;
  }
  wxCurrentEvent = NULL;
}


/************************************************************************
 * DIALOG AND PANEL ITEM FUNCTIONS
 *
 ************************************************************************/

// Create a new dialog box
long clipsDialogBoxCreate()
{
  if (ArgCountCheck("clipsDialogBoxCreate", AT_LEAST, 2) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *title = RtnLexeme(2);

  if (!title)
    return -1;

  // Set up defaults
  Bool modal = TRUE;
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;

  if (no_args > 2)
    modal = (Bool)RtnLong(3);

  if (no_args > 3)
    x = (int) RtnLong(4);

  if (no_args > 4)
    y = (int) RtnLong(5);

  if (no_args > 5)
    width = (int) RtnLong(6);

  if (no_args > 6)
    height = (int) RtnLong(7);

  wxFrame *frame = NULL;

  if (parent_id != 0)
  {
    frame = (wxFrame *)wxGetTypedObject(parent_id, wxTYPE_FRAME);
    if (!frame)
    {
      ClipsError << "Error in dialog-box-create: could not find parent frame " << parent_id << ".\n";
      return 0;
    }
  }

  wxClipsDialogBox *dialog_box = new wxClipsDialogBox(frame, title, modal, x, y, width, height);
  wxStoreThingInTable((long)dialog_box, dialog_box, wxTYPE_DIALOG_BOX);
  return (long)dialog_box;
}

// Create a new panel
long clipsPanelCreate()
{
  if (ArgCountCheck("clipsPanelCreate", AT_LEAST, 1) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);

  // Set up defaults
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;
  int style = 0;

  if (no_args > 1)
    x = (int) RtnLong(2);

  if (no_args > 2)
    y = (int) RtnLong(3);

  if (no_args > 3)
    width = (int) RtnLong(4);

  if (no_args > 4)
    height = (int) RtnLong(5);

  if (no_args > 5)
    style = (int) RtnLong(6);

  wxFrame *frame = NULL;

  if (parent_id != 0)
  {
    frame = (wxFrame *)wxGetTypedObject(parent_id, wxTYPE_FRAME);
    if (!frame)
    {
      ClipsError << "Error in panel-create: could not find parent frame " << parent_id << ".\n";
      return 0;
    }
  }

  wxClipsPanel *panel = new wxClipsPanel(frame, x, y, width, height, style);
  wxStoreThingInTable((long)panel, panel, wxTYPE_PANEL);
  return (long)panel;
}

// Set the label position
long clipsPanelSetLabelPosition()
{
  if (ArgCountCheck("clipsPanelSetLabelPosition", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *position = RtnLexeme(2);
  if (!position)
    return 0;

  wxPanel *panel = (wxPanel *)wxGetTypedObject(id, wxTYPE_PANEL);
  if (panel)
  {
    int pos;
    if (strcmp(position, "wxVERTICAL") == 0)
      pos = wxVERTICAL;
    else if (strcmp(position, "wxHORIZONTAL") == 0)
      pos = wxHORIZONTAL;
    else
    {
      ClipsError << "Error in panel-set-label-position: argument should be wxVERTICAL or wxHORIZONTAL.\n";
      return 0;
    }
    panel->SetLabelPosition(pos);
    return 1;
  }
  else
  {
    ClipsError << "Error in panel-set-label-position: could not find panel " << id << ".\n";
    return 0;
  }
}

// New line
long clipsPanelNewLine()
{
  if (ArgCountCheck("clipsPanelNewLine", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxPanel *panel = (wxPanel *)wxGetTypedObject(id, wxTYPE_PANEL);
  if (panel)
  {
    panel->NewLine();
    return 1;
  }
  else
  {
    ClipsError << "Error in panel-new-line: not find panel " << id << ".\n";
    return 0;
  }
}

// Create a new canvas
long clipsCanvasCreate()
{
  if (ArgCountCheck("clipsCanvasCreate", AT_LEAST, 1) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);

  // Set up defaults
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;
  int style = wxRETAINED;

  if (no_args > 1)
    x = (int) RtnLong(2);

  if (no_args > 2)
    y = (int) RtnLong(3);

  if (no_args > 3)
    width = (int) RtnLong(4);

  if (no_args > 4)
    height = (int) RtnLong(5);

  if (no_args > 5)
    style = (int) RtnLong(6);

  wxFrame *frame = NULL;

  if (parent_id != 0)
  {
    frame = (wxFrame *)wxGetTypedObject(parent_id, wxTYPE_FRAME);
    if (!frame)
    {
      ClipsError << "Error in canvas-create: could not find parent frame " << parent_id << ".\n";
      return 0;
    }
  }

  wxClipsCanvas *canvas = new wxClipsCanvas(frame, x, y, width, height, style);
  wxStoreThingInTable((long)canvas, canvas, wxTYPE_CANVAS);
  return (long)canvas;
}

// Set the canvas scrollbars
long clipsCanvasSetScrollbars()
{
  if (ArgCountCheck("clipsCanvasSetScrollbars", EXACTLY, 7) == -1)
    return 0;

  long id = RtnLong(1);
  long horiz = RtnLong(2);
  long vert = RtnLong(3);
  long x_length = RtnLong(4);
  long y_length = RtnLong(5);
  long x_page = RtnLong(6);
  long y_page = RtnLong(7);

  wxCanvas *canvas = (wxCanvas *)wxGetTypedObject(id, wxTYPE_CANVAS);
  if (canvas)
  {
    canvas->SetScrollbars((int)horiz, (int)vert,
                          (int)x_length, (int)y_length, (int)x_page, (int)y_page);
    return 1;
  }
  else
  {
    ClipsError << "Error in canvas-set-scrollbars: could not find canvas " << id << ".\n";
    return 0;
  }
}

// Set the canvas scrollbars
long clipsCanvasScroll()
{
  if (ArgCountCheck("clipsCanvasScroll", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  long x_pos = RtnLong(2);
  long y_pos = RtnLong(3);

  wxCanvas *canvas = (wxCanvas *)wxGetTypedObject(id, wxTYPE_CANVAS);
  if (canvas)
  {
    canvas->Scroll((int)x_pos, (int)y_pos);
    return 1;
  }
  else
  {
    ClipsError << "Error in canvas-scroll: could not find canvas " << id << ".\n";
    return 0;
  }
}

// Get canvas's device context
long clipsCanvasGetDC()
{
  if (ArgCountCheck("clipsCanvasGetDC", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxCanvas *canvas = (wxCanvas *)wxGetTypedObject(id, wxTYPE_CANVAS);
  if (canvas)
  {
    return (long)canvas->GetDC();
  }
  else
  {
    ClipsError << "Error in canvas-get-dc: could not find canvas " << id << ".\n";
    return 0;
  }
}

/*
 * Device context functions
 */

long clipsDCSetPen()
{
  if (ArgCountCheck("clipsDCSetPen", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long penId = RtnLong(2);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    wxPen *pen = (wxPen *)wxGetTypedObject(penId, wxTYPE_PEN);
    if (pen)
    {
      dc->SetPen(pen);
      return 1;
    }
    else
    {
      ClipsError << "Error in dc-set-pen: could not find pen " << penId << ".\n";
      return 0;
    }
  }
  else
  {
    ClipsError << "Error in dc-set-pen: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCSetBrush()
{
  if (ArgCountCheck("clipsDCSetBrush", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long brushId = RtnLong(2);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    wxBrush *brush = (wxBrush *)wxGetTypedObject(brushId, wxTYPE_BRUSH);
    if (brush)
    {
      dc->SetBrush(brush);
      return 1;
    }
    else
    {
      ClipsError << "Error in dc-set-brush: could not find brush " << brushId << ".\n";
      return 0;
    }
  }
  else
  {
    ClipsError << "Error in dc-set-brush: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCSetFont()
{
  if (ArgCountCheck("clipsDCSetFont", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long fontId = RtnLong(2);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    wxFont *font = (wxFont *)wxGetTypedObject(fontId, wxTYPE_FONT);
    if (font)
    {
      dc->SetFont(font);
      return 1;
    }
    else
    {
      ClipsError << "Error in dc-set-font: could not find font " << fontId << ".\n";
      return 0;
    }
  }
  else
  {
    ClipsError << "Error in dc-set-font: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCSetTextForeground()
{
  if (ArgCountCheck("clipsDCSetTextForeground", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *colour = RtnLexeme(2);
  if (!colour)
    return 0;

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    wxColour *col = wxTheColourDatabase->FindColour(colour);
    if (col)
    {
      dc->SetTextForeground(col);
      return 1;
    }
    else
    {
      ClipsError << "Error in dc-set-text-foreground: could not find colour " << colour << ".\n";
      return 0;
    }
  }
  else
  {
    ClipsError << "Error in dc-set-text-foreground: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCSetTextBackground()
{
  if (ArgCountCheck("clipsDCSetTextBackground", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *colour = RtnLexeme(2);
  if (!colour)
    return 0;

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    wxColour *col = wxTheColourDatabase->FindColour(colour);
    if (col)
    {
      dc->SetTextBackground(col);
      return 1;
    }
    else
    {
      ClipsError << "Error in dc-set-text-background: could not find colour " << colour << ".\n";
      return 0;
    }
  }
  else
  {
    ClipsError << "Error in dc-set-text-background: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCSetLogicalFunction()
{
  if (ArgCountCheck("clipsDCSetLogicalFunction", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *op = RtnLexeme(2);
  if (!op)
    return 0;

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    int log_op = 0;
    
    if (strcmp(op, "wxCOPY") == 0)
      log_op = wxCOPY;
    else if (strcmp(op, "wxXOR") == 0)
      log_op = wxXOR;
    else if (strcmp(op, "wxINVERT") == 0)
      log_op = wxINVERT;
    else if (strcmp(op, "wxOR_REVERSE") == 0)
      log_op = wxOR_REVERSE;
    else if (strcmp(op, "wxAND_REVERSE") == 0)
      log_op = wxAND_REVERSE;
    else
    {
      ClipsError << "Error in dc-set-logical-function: unrecognized function " << op << "\n";
      ClipsError << "Function must be one of wxCOPY, wxXOR, wxINVERT, wxOR_REVERSE, wxAND_REVERSE.\n";
      return 0;
    }
    dc->SetLogicalFunction(log_op);
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-set-logical-function: could not find dc " << id << ".\n";
    return 0;
  }
}
 
long clipsDCDrawLine()
{
  if (ArgCountCheck("clipsDCDrawLine", EXACTLY, 5) == -1)
    return 0;

  long id = RtnLong(1);
  float x1 = (float)RtnDouble(2);
  float y1 = (float)RtnDouble(3);
  float x2 = (float)RtnDouble(4);
  float y2 = (float)RtnDouble(5);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    dc->DrawLine(x1, y1, x2, y2);
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-draw-line: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCDrawPoint()
{
  if (ArgCountCheck("clipsDCDrawPoint", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  float x1 = (float)RtnDouble(2);
  float y1 = (float)RtnDouble(3);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    dc->DrawPoint(x1, y1);
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-draw-point: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCDrawRectangle()
{
  if (ArgCountCheck("clipsDCDrawRectangle", EXACTLY, 5) == -1)
    return 0;

  long id = RtnLong(1);
  float x1 = (float)RtnDouble(2);
  float y1 = (float)RtnDouble(3);
  float x2 = (float)RtnDouble(4);
  float y2 = (float)RtnDouble(5);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    dc->DrawRectangle(x1, y1, x2, y2);
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-draw-rectangle: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCDrawRoundedRectangle()
{
  if (ArgCountCheck("clipsDCDrawRoundedRectangle", AT_LEAST, 5) == -1)
    return 0;

  long id = RtnLong(1);
  float x1 = (float)RtnDouble(2);
  float y1 = (float)RtnDouble(3);
  float x2 = (float)RtnDouble(4);
  float y2 = (float)RtnDouble(5);
  float radius = 20;
  if (RtnArgCount() > 5)
    radius = (float)RtnDouble(6);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    dc->DrawRoundedRectangle(x1, y1, x2, y2, radius);
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-draw-rounded-rectangle: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCDrawEllipse()
{
  if (ArgCountCheck("clipsDCDrawEllipse", EXACTLY, 5) == -1)
    return 0;

  long id = RtnLong(1);
  float x = (float)RtnDouble(2);
  float y = (float)RtnDouble(3);
  float width = (float)RtnDouble(4);
  float height = (float)RtnDouble(5);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    dc->DrawEllipse(x, y, width, height);
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-draw-ellipse: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCDrawSpline()
{
  DATA_OBJECT arg;

  if (ArgCountCheck("clipsDCDrawSpline", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    RtnUnknown(2, &arg);

    if (GetType(arg) != MULTIFIELD)
    {
      ClipsError << "Error in dc-draw-spline: 2nd argument must be a multifield of floating-point numbers.\n";
      return 0;
    }

    int start = GetpDOBegin(&arg);
    int end = GetpDOEnd(&arg);

    VOID *multifieldPtr = ClipsGetValue(arg);

    int i;
    wxList pointList;
    for (i = start; i <= end; i += 2)
    {
      double x, y;
      if ((GetMFType(multifieldPtr, i) == FLOAT) || (GetMFType(multifieldPtr, i) == INTEGER))
        x = ValueToDouble(GetMFValue(multifieldPtr, i));
      else
      {
        ClipsError << "Error in dc-draw-spline: multifield element " << i-start+1 << " must be a number.\n";
        return 0;
      }
      if (((i+1) <= end) && ((GetMFType(multifieldPtr, i+1) == FLOAT) || (GetMFType(multifieldPtr, i+1) == INTEGER)))
        y = ValueToDouble(GetMFValue(multifieldPtr, i+1));
      else
      {
        ClipsError << "Error in dc-draw-spline: multifield element " << i-start+2 << " must be a number.\n";
        return 0;
      }
      pointList.Append(new wxPoint((float)x, (float)y));
    }
    dc->DrawSpline(&pointList);
    wxNode *node = pointList.First();
    while (node)
    {
      wxPoint *point = (wxPoint *)node->Data();
      delete point;
      node = node->Next();
    }
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-draw-spline: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCDrawPolygon()
{
  DATA_OBJECT arg;

  if (ArgCountCheck("clipsDCDrawPolygon", AT_LEAST, 2) == -1)
    return 0;

  float x_offset = 0.0;
  float y_offset = 0.0;
  
  long id = RtnLong(1);
  int no_args = RtnArgCount();
  if (no_args > 2)
    x_offset = (float)RtnDouble(3);
  if (no_args > 3)
    y_offset = (float)RtnDouble(4);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    RtnUnknown(2, &arg);

    if (GetType(arg) != MULTIFIELD)
    {
      ClipsError << "Error in dc-draw-polygon: 2nd argument must be a multifield of floating-point numbers.\n";
      return 0;
    }

    int start = GetpDOBegin(&arg);
    int end = GetpDOEnd(&arg);

    VOID *multifieldPtr = ClipsGetValue(arg);

    int i;
    wxList pointList;
    for (i = start; i <= end; i += 2)
    {
      double x, y;
      if ((GetMFType(multifieldPtr, i) == FLOAT) || (GetMFType(multifieldPtr, i) == INTEGER))
        x = ValueToDouble(GetMFValue(multifieldPtr, i));
      else
      {
        ClipsError << "Error in dc-draw-polygon: multifield element " << i-start+1 << " must be a number.\n";
        return 0;
      }
      if (((i+1) <= end) && ((GetMFType(multifieldPtr, i+1) == FLOAT) || (GetMFType(multifieldPtr, i+1) == INTEGER)))
        y = ValueToDouble(GetMFValue(multifieldPtr, i+1));
      else
      {
        ClipsError << "Error in dc-draw-polygon: multifield element " << i-start+2 << " must be a number.\n";
        return 0;
      }
      pointList.Append(new wxPoint((float)x, (float)y));
    }
    dc->DrawPolygon(&pointList);
    wxNode *node = pointList.First();
    while (node)
    {
      wxPoint *point = (wxPoint *)node->Data();
      delete point;
      node = node->Next();
    }
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-draw-polygon: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCDrawLines()
{
  DATA_OBJECT arg;

  if (ArgCountCheck("clipsDCDrawLines", AT_LEAST, 2) == -1)
    return 0;

  float x_offset = 0.0;
  float y_offset = 0.0;
  
  long id = RtnLong(1);
  int no_args = RtnArgCount();
  if (no_args > 2)
    x_offset = (float)RtnDouble(3);
  if (no_args > 3)
    y_offset = (float)RtnDouble(4);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    RtnUnknown(2, &arg);

    if (GetType(arg) != MULTIFIELD)
    {
      ClipsError << "Error in dc-draw-lines: 2nd argument must be a multifield of floating-point numbers.\n";
      return 0;
    }

    int start = GetpDOBegin(&arg);
    int end = GetpDOEnd(&arg);

    VOID *multifieldPtr = ClipsGetValue(arg);

    int i;
    wxList pointList;
    for (i = start; i <= end; i += 2)
    {
      double x, y;
      if ((GetMFType(multifieldPtr, i) == FLOAT) || (GetMFType(multifieldPtr, i) == INTEGER))
        x = ValueToDouble(GetMFValue(multifieldPtr, i));
      else
      {
        ClipsError << "Error in dc-draw-lines: multifield element " << i-start+1 << " must be a number.\n";
        return 0;
      }
      if (((i+1) <= end) && ((GetMFType(multifieldPtr, i+1) == FLOAT) || (GetMFType(multifieldPtr, i+1) == INTEGER)))
        y = ValueToDouble(GetMFValue(multifieldPtr, i+1));
      else
      {
        ClipsError << "Error in dc-draw-lines: multifield element " << i-start+2 << " must be a number.\n";
        return 0;
      }
      pointList.Append(new wxPoint((float)x, (float)y));
    }
    dc->DrawPolygon(&pointList);
    wxNode *node = pointList.First();
    while (node)
    {
      wxPoint *point = (wxPoint *)node->Data();
      delete point;
      node = node->Next();
    }
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-draw-lines: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCDrawText()
{
  if (ArgCountCheck("clipsDCDrawText", EXACTLY, 4) == -1)
    return 0;

  long id = RtnLong(1);
  char *text = RtnLexeme(2);
  float x1 = (float)RtnDouble(3);
  float y1 = (float)RtnDouble(4);
  if (!text)
    return FALSE;

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    dc->DrawText(text, x1, y1);
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-draw-text: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCSetClippingRegion()
{
  if (ArgCountCheck("clipsDCSetClippingRegion", EXACTLY, 5) == -1)
    return 0;

  long id = RtnLong(1);
  float x1 = (float)RtnDouble(2);
  float y1 = (float)RtnDouble(3);
  float x2 = (float)RtnDouble(4);
  float y2 = (float)RtnDouble(5);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    dc->SetClippingRegion(x1, y1, x2, y2);
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-set-clipping-region: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCDestroyClippingRegion()
{
  if (ArgCountCheck("clipsDCDestroyClippingRegion", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    dc->DestroyClippingRegion();
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-destroy-clipping-region: could not find dc " << id << ".\n";
    return 0;
  }
}

long clipsDCClear()
{
  if (ArgCountCheck("clipsDCClear", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxDC *dc = (wxDC *)wxGetTypedObject(id, wxTYPE_DC);
  if (dc)
  {
    dc->Clear();
    return 1;
  }
  else
  {
    ClipsError << "Error in dc-clear: could not find dc " << id << ".\n";
    return 0;
  }
}

// Create a new pen
long clipsPenCreate()
{
  if (ArgCountCheck("clipsPenCreate", EXACTLY, 3) == -1)
    return 0;

  char *colour = RtnLexeme(1);
  long width = RtnLong(2);
  char *style = RtnLexeme(3);
  if (!colour || !style)
    return -1;

  wxColour *col = wxTheColourDatabase->FindColour(colour);
  int sty = 0;
  if (!col)
  {
    ClipsError << "Error in pen-create: could not find colour " << colour << ".\n";
    return 0;
  }
  if (strcmp(style, "wxSOLID") == 0)
    sty = wxSOLID;
  else if (strcmp(style, "wxDOT") == 0)
    sty = wxDOT;
  else if (strcmp(style, "wxLONG_DASH") == 0)
    sty = wxLONG_DASH;
  else if (strcmp(style, "wxSHORT_DASH") == 0)
    sty = wxSHORT_DASH;
  else if (strcmp(style, "wxDOT_DASH") == 0)
    sty = wxDOT_DASH;
  else if (strcmp(style, "wxTRANSPARENT") == 0)
    sty = wxTRANSPARENT;
  else
  {
    ClipsError << "Error in pen-create: unrecognized style " << style << "\n";
    ClipsError << "Style must be one of wxSOLID, wxDOT, wxLONG_DASH, wxSHORT_DASH, wxDOT_DASH, wxTRANSPARENT.\n";
    return 0;
  }

  wxClipsPen *pen = new wxClipsPen(*col, (int)width, sty);
  wxStoreThingInTable((long)pen, pen, wxTYPE_PEN);
  return (long)pen;
}

// Delete pen
long clipsPenDelete()
{
  if (ArgCountCheck("clipsPenDelete", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  wxPen *pen = (wxPen *)wxGetTypedObject(id, wxTYPE_PEN);
  if (pen)
  {
    delete pen;
    return 1;
  }
  else
  {
    ClipsError << "Error in pen-delete: could not find pen " << id << ".\n";
    return 0;
  }
}

// Create a new brush
long clipsBrushCreate()
{
  if (ArgCountCheck("clipsBrushCreate", EXACTLY, 2) == -1)
    return 0;

  char *colour = RtnLexeme(1);
  char *style = RtnLexeme(2);
  if (!colour || !style)
    return -1;

  wxColour *col = wxTheColourDatabase->FindColour(colour);
  int sty = 0;
  if (!col)
  {
    ClipsError << "Error in brush-create: could not find colour " << colour << ".\n";
    return 0;
  }
  if (strcmp(style, "wxSOLID") == 0)
    sty = wxSOLID;
  else if (strcmp(style, "wxTRANSPARENT") == 0)
    sty = wxTRANSPARENT;
  else
  {
    ClipsError << "Error in brush-create: unrecognized style " << style << "\n";
    ClipsError << "Style must be one of wxSOLID, wxTRANSPARENT.\n";
    return 0;
  }

  wxClipsBrush *brush = new wxClipsBrush(*col, sty);
  wxStoreThingInTable((long)brush, brush, wxTYPE_BRUSH);
  return (long)brush;
}

// Delete pen
long clipsBrushDelete()
{
  if (ArgCountCheck("clipsBrushDelete", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  wxBrush *brush = (wxBrush *)wxGetTypedObject(id, wxTYPE_BRUSH);
  if (brush)
  {
    delete brush;
    return 1;
  }
  else
  {
    ClipsError << "Error in brush-delete: could not find brush " << id << ".\n";
    return 0;
  }
}

// Create a new font
long clipsFontCreate()
{
  if (ArgCountCheck("clipsFontCreate", EXACTLY, 5) == -1)
    return 0;

  int pointSize = (int)RtnLong(1);
  char *family = RtnLexeme(2);
  char *style = RtnLexeme(3);
  char *weight  = RtnLexeme(4);
  int wxStyle = 0;
  int wxFamily = 0;
  int wxWeight = 0;
  Bool underlined = (Bool)RtnLong(5);
  if (!family || !style || !weight)
    return -1;

  if (strcmp(family, "wxROMAN") == 0)
    wxFamily = wxROMAN;
  else if (strcmp(family, "wxSCRIPT") == 0)
    wxFamily = wxSCRIPT;
  else if (strcmp(family, "wxDECORATIVE") == 0)
    wxFamily = wxDECORATIVE;
  else if (strcmp(family, "wxSWISS") == 0)
    wxFamily = wxSWISS;
  else if (strcmp(family, "wxMODERN") == 0)
    wxFamily = wxMODERN;
  else if (strcmp(family, "wxDEFAULT") == 0)
    wxFamily = wxMODERN;
  else
  {
    ClipsError << "Error in font-create: unrecognized family " << family << ".\n";
    ClipsError << "Family must be one of wxDECORATIVE, wxDEFAULT, wxMODERN, wxROMAN, wxSCRIPT, wxSWISS.\n";
    return 0;
  }

  if (strcmp(weight, "wxNORMAL") == 0)
    wxStyle = wxNORMAL;
  else if (strcmp(weight, "wxBOLD") == 0)
    wxStyle = wxBOLD;
  else if (strcmp(weight, "wxLIGHT") == 0)
    wxStyle = wxLIGHT;
  else
  {
    ClipsError << "Error in font-create: unrecognized weight " << weight << ".\n";
    ClipsError << "Weight must be one of wxNORMAL, wxBOLD, wxLIGHT.\n";
    return 0;
  }

  if (strcmp(style, "wxNORMAL") == 0)
    wxStyle = wxNORMAL;
  else if (strcmp(style, "wxITALIC") == 0)
    wxStyle = wxITALIC;
  else
  {
    ClipsError << "Error in font-create: unrecognized style " << style << ".\n";
    ClipsError << "Style must be one of wxNORMAL, wxITALIC.\n";
    return 0;
  }

  wxClipsFont *font = new wxClipsFont(pointSize, wxFamily, wxStyle, wxWeight);
  wxStoreThingInTable((long)font, font, wxTYPE_FONT);
  return (long)font;
}

// Delete font
long clipsFontDelete()
{
  if (ArgCountCheck("clipsFontDelete", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  wxFont *font = (wxFont *)wxGetTypedObject(id, wxTYPE_FONT);
  if (font)
  {
    delete font;
    return 1;
  }
  else
  {
    ClipsError << "Error in font-delete: could not find font " << id << ".\n";
    return 0;
  }
}

// Create a new icon
long clipsIconCreate()
{
  if (ArgCountCheck("clipsIconCreate", EXACTLY, 1) == -1)
    return 0;

  char *file = RtnLexeme(1);
  if (!file)
    return -1;

  wxClipsIcon *icon = new wxClipsIcon(file);
  wxStoreThingInTable((long)icon, icon, wxTYPE_ICON);
  return (long)icon;
}

// Delete icon
long clipsIconDelete()
{
  if (ArgCountCheck("clipsIconDelete", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  wxIcon *icon = (wxIcon *)wxGetTypedObject(id, wxTYPE_ICON);
  if (icon)
  {
    delete icon;
    return 1;
  }
  else
  {
    ClipsError << "Error in icon-delete: could not find icon " << id << ".\n";
    return 0;
  }
}

/*
 * Event accessors
 */

long clipsMouseEventIsButton()
{
  if (ArgCountCheck("clipsMouseEventIsButton", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->IsButton();
  }
  else
  {
    ClipsError << "Error in event-is-button: could not find event " << id << ".\n";
    return 0;
  }
}

long clipsMouseEventButtonDown()
{
  if (ArgCountCheck("clipsMouseEventButtonDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->ButtonDown();
  }
  else
  {
    ClipsError << "Error in event-button-down: could not find event " << id << ".\n";
    return 0;
  }
}

long clipsMouseEventControlDown()
{
  if (ArgCountCheck("clipsMouseEventControlDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->ControlDown();
  }
  else
  {
    ClipsError << "Error in event-control-down: could not find event " << id << ".\n";
    return 0;
  }
}

long clipsMouseEventShiftDown()
{
  if (ArgCountCheck("clipsMouseEventShiftDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->ShiftDown();
  }
  else
  {
    ClipsError << "Error in event-shift-down: could not find event " << id << ".\n";
    return 0;
  }
}

long clipsMouseEventButton()
{
  if (ArgCountCheck("clipsMouseEventButton", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long butt = RtnLong(2);
  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->Button((int)butt);
  }
  else
  {
    ClipsError << "Error in event-button: could not find event " << id << ".\n";
    return 0;
  }
}

long clipsMouseEventLeftDown()
{
  if (ArgCountCheck("clipsMouseEventLeftDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->LeftDown();
  }
  else
  {
    ClipsError << "Error in event-left-down: could not find event " << id << ".\n";
    return 0;
  }
}

long clipsMouseEventMiddleDown()
{
  if (ArgCountCheck("clipsMouseEventMiddleDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->MiddleDown();
  }
  else
  {
    ClipsError << "Error in event-middle-down: could not find event " << id << ".\n";
    return 0;
  }
}

long clipsMouseEventRightDown()
{
  if (ArgCountCheck("clipsMouseEventRightDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->RightDown();
  }
  else
  {
    ClipsError << "Error in event-right-down: could not find event " << id << ".\n";
    return 0;
  }
}

long clipsMouseEventLeftUp()
{
  if (ArgCountCheck("clipsMouseEventLeftUp", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->LeftUp();
  }
  else
  {
    ClipsError << "Error in event-left-up: could not find event " << id << ".\n";
    return 0;
  }
}

long clipsMouseEventMiddleUp()
{
  if (ArgCountCheck("clipsMouseEventMiddleUp", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->MiddleUp();
  }
  else
  {
    ClipsError << "Error in event-middle-up: could not find event " << id << ".\n";
    return 0;
  }
}

long clipsMouseEventRightUp()
{
  if (ArgCountCheck("clipsMouseEventRightUp", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->RightUp();
  }
  else
  {
    ClipsError << "Error in event-right-up: could not find event " << id << ".\n";
    return 0;
  }
}

long clipsMouseEventDragging()
{
  if (ArgCountCheck("clipsMouseEventDragging", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    return event->Dragging();
  }
  else
  {
    ClipsError << "Error in event-dragging: could not find event " << id << ".\n";
    return 0;
  }
}

double clipsMouseEventPositionX()
{
  if (ArgCountCheck("clipsMouseEventPositionX", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    float x, y;
    event->Position(&x, &y);
    return (double)x;
  }
  else
  {
    ClipsError << "Error in event-position-x: could not find event " << id << ".\n";
    return 0.0;
  }
}

double clipsMouseEventPositionY()
{
  if (ArgCountCheck("clipsMouseEventPositionY", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxMouseEvent *event = (wxMouseEvent *)wxGetTypedObject(id, wxTYPE_MOUSE_EVENT);
  if (event)
  {
    float x, y;
    event->Position(&x, &y);
    return (double)y;
  }
  else
  {
    ClipsError << "Error in event-position-y: could not find event " << id << ".\n";
    return 0.0;
  }
}


// Create a new panel
long clipsTextWindowCreate()
{
  if (ArgCountCheck("clipsTextWindowCreate", AT_LEAST, 1) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);

  // Set up defaults
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;
  int style = 0;

  if (no_args > 1)
    x = (int) RtnLong(2);

  if (no_args > 2)
    y = (int) RtnLong(3);

  if (no_args > 3)
    width = (int) RtnLong(4);

  if (no_args > 4)
    height = (int) RtnLong(5);

  if (no_args > 5)
    style = (int) RtnLong(6);

  wxFrame *frame = NULL;

  if (parent_id != 0)
  {
    frame = (wxFrame *)wxGetTypedObject(parent_id, wxTYPE_FRAME);
    if (!frame)
    {
      ClipsError << "Error in text-window-create: could not find parent frame " << parent_id << ".\n";
      return 0;
    }
  }

  wxClipsTextWindow *text_window = new wxClipsTextWindow(frame, x, y, width, height, style);
  wxStoreThingInTable((long)text_window, text_window, wxTYPE_TEXT_WINDOW);
  return (long)text_window;
}

// Load a file
long clipsTextWindowLoadFile()
{
  if (ArgCountCheck("clipsTextWindowLoadFile", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *filename = RtnLexeme(2);
  if (!filename)
    return 0;

  wxTextWindow *win = (wxTextWindow *)wxGetTypedObject(id, wxTYPE_TEXT_WINDOW);
  if (win)
  {
    return (long)win->LoadFile(filename);
  }
  else
  {
    ClipsError << "Error in text-window-load-file: could not find text window " << id << ".\n";
    return 0;
  }
}

// Save a file
long clipsTextWindowSaveFile()
{
  if (ArgCountCheck("clipsTextWindowSaveFile", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *filename = RtnLexeme(2);
  if (!filename)
    return 0;

  wxTextWindow *win = (wxTextWindow *)wxGetTypedObject(id, wxTYPE_TEXT_WINDOW);
  if (win)
  {
    return (long)win->SaveFile(filename);
  }
  else
  {
    ClipsError << "Error in text-window-save-file: could not find text window " << id << ".\n";
    return 0;
  }
}

// Write text to the window
VOID clipsTextWindowWrite()
{
  if (ArgCountCheck("clipsTextWindowWrite", AT_LEAST, 1) == -1)
    return;

  long id = RtnLong(1);

  wxTextWindow *win = (wxTextWindow *)wxGetTypedObject(id, wxTYPE_TEXT_WINDOW);
  if (win)
  {
    DATA_OBJECT temp;

    for (int i = 2; i <= RtnArgCount(); i++)
    {
      RtnUnknown(i, &temp);

      switch (GetType(temp))
      {
        case INTEGER:
        {
          *win << DOToLong(temp);
          break;
        }
        case STRING:
        case SYMBOL:
        {
          *win << DOToString(temp);
          break;
        }
        case FLOAT:
        {
          *win << DOToDouble(temp);
          break;
        }
        default:
        {
          ClipsError << "Error in text-window-write: unrecognized type.\n";
          break;
        }
      }
    }
    return;
  }
  else
  {
    ClipsError << "Error in text-window-write: could not find text window " << id << ".\n";
    return;
  }
}

// Return TRUE if modified
long clipsTextWindowModified()
{
  if (ArgCountCheck("clipsTextWindowModified", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxTextWindow *win = (wxTextWindow *)wxGetTypedObject(id, wxTYPE_TEXT_WINDOW);
  if (win)
  {
    return (long)win->Modified();
  }
  else
  {
    ClipsError << "Error in text-window-modified: could not find text window " << id << ".\n";
    return 0;
  }
}

// Return TRUE if modified
VOID clipsTextWindowDiscardEdits()
{
  if (ArgCountCheck("clipsTextWindowDiscardEdits", EXACTLY, 1) == -1)
    return;

  long id = RtnLong(1);

  wxTextWindow *win = (wxTextWindow *)wxGetTypedObject(id, wxTYPE_TEXT_WINDOW);
  if (win)
  {
    win->DiscardEdits();
    return;
  }
  else
  {
    ClipsError << "Error in text-window-discard-edits: could not find text window " << id << ".\n";
    return;
  }
}

// Create a button
long clipsButtonCreate()
{
  if (ArgCountCheck("clipsButtonCreate", AT_LEAST, 3) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *func = RtnLexeme(2);

  if (func && (strlen(func) == 0))
    func = NULL;

  char *label = RtnLexeme(3);
  if (!label)
    return -1;

  // Set up defaults
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;

  if (no_args > 3)
    x = (int) RtnLong(4);

  if (no_args > 4)
    y = (int) RtnLong(5);

  if (no_args > 5)
    width = (int) RtnLong(6);

  if (no_args > 6)
    height = (int) RtnLong(7);

  wxPanel *panel = (wxPanel *)wxGetTypedObject(parent_id, wxTYPE_PANEL);
  if (panel)
  {
    wxClipsButton *button = new wxClipsButton(panel, func, (wxFunction)wxClipsButtonCallback, label, x, y, width, height);
    wxStoreThingInTable((long)button, button, wxTYPE_BUTTON);
    return (long)button;
  }
  else
  {
    ClipsError << "Error in button-create: could not find parent panel " << parent_id << ".\n";
    return 0;
  }
}

// Create a text widget
long clipsTextCreate()
{
  if (ArgCountCheck("clipsTextCreate", AT_LEAST, 3) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *func = RtnLexeme(2);

  if (func && (strlen(func) == 0))
    func = NULL;

  char *label = RtnLexeme(3);
  if (!label)
    return -1;

  // Set up defaults
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;
  char *value = "";

  if (no_args > 3)
    value = RtnLexeme(4);
  if (!value)
    return -1;

  if (no_args > 4)
    x = (int) RtnLong(5);

  if (no_args > 5)
    y = (int) RtnLong(6);

  if (no_args > 6)
    width = (int) RtnLong(7);

  if (no_args > 7)
    height = (int) RtnLong(8);

  wxPanel *panel = (wxPanel *)wxGetTypedObject(parent_id, wxTYPE_PANEL);
  if (panel)
  {
    wxClipsText *text = new wxClipsText(panel, func, (wxFunction)wxClipsTextCallback, label, value, x, y, width, height);
    wxStoreThingInTable((long)text, text, wxTYPE_TEXT);
    return (long)text;
  }
  else
  {
    ClipsError << "Error in text-create: could not find parent panel " << parent_id << ".\n";
    return 0;
  }
}

// Set value
long clipsTextSetValue()
{
  if (ArgCountCheck("clipsTextSetValue", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *val = RtnLexeme(2);
  if (!val)
    return 0;

  wxText *item = (wxText *)wxGetTypedObject(id, wxTYPE_TEXT);
  if (item)
  {
    item->SetValue(val);
    return 1;
  }
  else
  {
    ClipsError << "Error in text-set-value: could not find text item " << id << ".\n";
    return 0;
  }
}

// Get value
char *clipsTextGetValue()
{
  if (ArgCountCheck("clipsTextGetValue", EXACTLY, 1) == -1)
    return (char *)AddSymbol("");

  long id = RtnLong(1);

  wxText *item = (wxText *)wxGetTypedObject(id, wxTYPE_TEXT);
  if (item)
  {
    char *val = item->GetValue();
    return (char *)AddSymbol(val);
  }
  else
  {
    ClipsError << "Error in text-get-value: could not find text item " << id << ".\n";
    return (char *)AddSymbol("");
  }
}

// Create a multi-line text widget
long clipsMultiTextCreate()
{
  if (ArgCountCheck("clipsMultiTextCreate", AT_LEAST, 3) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *func = RtnLexeme(2);
  if (func && (strlen(func) == 0))
    func = NULL;

  char *label = RtnLexeme(3);
  if (!label)
    return -1;

  // Set up defaults
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;
  char *value = "";

  if (no_args > 3)
    value = RtnLexeme(4);
  if (!value)
    return -1;

  if (no_args > 4)
    x = (int) RtnLong(5);

  if (no_args > 5)
    y = (int) RtnLong(6);

  if (no_args > 6)
    width = (int) RtnLong(7);

  if (no_args > 7)
    height = (int) RtnLong(8);

  wxPanel *panel = (wxPanel *)wxGetTypedObject(parent_id, wxTYPE_PANEL);
  if (panel)
  {
    wxClipsMultiText *text = new wxClipsMultiText(panel, func, (wxFunction)wxClipsMultiTextCallback, label, value, x, y, width, height);
    wxStoreThingInTable((long)text, text, wxTYPE_MULTI_TEXT);
    return (long)text;
  }
  else
  {
    ClipsError << "Error in text-create: could not find parent panel " << parent_id << ".\n";
    return 0;
  }
}

// Set value
long clipsMultiTextSetValue()
{
  if (ArgCountCheck("clipsMultiTextSetValue", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *val = RtnLexeme(2);
  if (!val)
    return 0;

  wxMultiText *item = (wxMultiText *)wxGetTypedObject(id, wxTYPE_MULTI_TEXT);
  if (item)
  {
    item->SetValue(val);
    return 1;
  }
  else
  {
    ClipsError << "Error in multi-text-set-value: could not find text item " << id << ".\n";
    return 0;
  }
}

// Get value
char *clipsMultiTextGetValue()
{
  if (ArgCountCheck("clipsMultiTextGetValue", EXACTLY, 1) == -1)
    return (char *)AddSymbol("");

  long id = RtnLong(1);

  wxMultiText *item = (wxMultiText *)wxGetTypedObject(id, wxTYPE_MULTI_TEXT);
  if (item)
  {
    char *val = item->GetValue();
    return (char *)AddSymbol(val);
  }
  else
  {
    ClipsError << "Error in multi-text-get-value: could not find text item " << id << ".\n";
    return (char *)AddSymbol("");
  }
}


// Create a check box widget
long clipsCheckBoxCreate()
{
  if (ArgCountCheck("clipsCheckBoxCreate", AT_LEAST, 2) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *func = RtnLexeme(2);
  if (func && (strlen(func) == 0))
    func = NULL;

  char *label = RtnLexeme(3);
  if (!label)
    return -1;

  // Set up defaults
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;

  if (no_args > 3)
    x = (int) RtnLong(4);

  if (no_args > 4)
    y = (int) RtnLong(5);

  if (no_args > 5)
    width = (int) RtnLong(6);

  if (no_args > 6)
    height = (int) RtnLong(7);

  wxPanel *panel = (wxPanel *)wxGetTypedObject(parent_id, wxTYPE_PANEL);
  if (panel)
  {
    wxClipsCheckBox *box = new wxClipsCheckBox(panel, func, (wxFunction)wxClipsCheckBoxCallback, label, x, y, width, height);
    wxStoreThingInTable((long)box, box, wxTYPE_CHECK_BOX);
    return (long)box;
  }
  else
  {
    ClipsError << "Error in check-box-create: could not find parent panel " << parent_id << ".\n";
    return 0;
  }
}

// Set value
long clipsCheckBoxSetValue()
{
  if (ArgCountCheck("clipsCheckBoxSetValue", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  Bool val = (Bool)RtnLong(2);

  wxCheckBox *item = (wxCheckBox *)wxGetTypedObject(id, wxTYPE_CHECK_BOX);
  if (item)
  {
    item->SetValue(val);
    return 1;
  }
  else
  {
    ClipsError << "Error in check-box-set-value: could not find check box " << id << ".\n";
    return 0;
  }
}

// Get value
long clipsCheckBoxGetValue()
{
  if (ArgCountCheck("clipsCheckBoxGetValue", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxCheckBox *item = (wxCheckBox *)wxGetTypedObject(id, wxTYPE_CHECK_BOX);
  if (item)
  {
    Bool val = item->GetValue();
    return val;
  }
  else
  {
    ClipsError << "Error in check-box-get-value: could not find check box " << id << ".\n";
    return -1;
  }
}

// Create a slider widget
long clipsSliderCreate()
{
  if (ArgCountCheck("clipsSliderCreate", AT_LEAST, 7) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *func = RtnLexeme(2);
  if (func && (strlen(func) == 0))
    func = NULL;

  char *label = RtnLexeme(3);
  if (!label)
    return -1;

  int value = (int)RtnLong(4);
  int min_value = (int)RtnLong(5);
  int max_value = (int)RtnLong(6);
  int width = (int)RtnLong(7);

  // Set up defaults
  int x = -1;
  int y = -1;

  if (no_args > 7)
    x = (int) RtnLong(8);

  if (no_args > 8)
    y = (int) RtnLong(9);

  wxPanel *panel = (wxPanel *)wxGetTypedObject(parent_id, wxTYPE_PANEL);
  if (panel)
  {
    wxClipsSlider *slider = new wxClipsSlider(panel, func, (wxFunction)wxClipsSliderCallback, label, value, min_value, max_value, width, x, y);
    wxStoreThingInTable((long)slider, slider, wxTYPE_SLIDER);
    return (long)slider;
  }
  else
  {
    ClipsError << "Error in slider-create: could not find parent panel " << parent_id << ".\n";
    return 0;
  }
}

// Set value
long clipsSliderSetValue()
{
  if (ArgCountCheck("clipsSliderSetValue", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  int val = (int)RtnLong(2);

  wxSlider *item = (wxSlider *)wxGetTypedObject(id, wxTYPE_SLIDER);
  if (item)
  {
    item->SetValue(val);
    return 1;
  }
  else
  {
    ClipsError << "Error in slider-set-value: could not find slider " << id << ".\n";
    return 0;
  }
}

// Get value
long clipsSliderGetValue()
{
  if (ArgCountCheck("clipsSliderGetValue", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxSlider *item = (wxSlider *)wxGetTypedObject(id, wxTYPE_SLIDER);
  if (item)
  {
    int val = item->GetValue();
    return (long)val;
  }
  else
  {
    ClipsError << "Error in slider-get-value: could not find slider " << id << ".\n";
    return -1;
  }
}

// Create a message widget
long clipsMessageCreate()
{
  if (ArgCountCheck("clipsMessageCreate", AT_LEAST, 2) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *label = RtnLexeme(2);
  if (!label)
    return -1;

  // Set up defaults
  int x = -1;
  int y = -1;

  if (no_args > 2)
    x = (int) RtnLong(3);

  if (no_args > 3)
    y = (int) RtnLong(4);

  wxPanel *panel = (wxPanel *)wxGetTypedObject(parent_id, wxTYPE_PANEL);
  if (panel)
  {
    wxClipsMessage *mess = new wxClipsMessage(panel, label, x, y);
    wxStoreThingInTable((long)mess, mess, wxTYPE_MESSAGE);
    return (long)mess;
  }
  else
  {
    ClipsError << "Error in message-create: could not find parent panel " << parent_id << ".\n";
    return 0;
  }
}

// Create a listbox widget
long clipsListBoxCreate()
{
  if (ArgCountCheck("clipsListBoxCreate", AT_LEAST, 3) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *func = RtnLexeme(2);
  if (func && (strlen(func) == 0))
    func = NULL;

  char *label = RtnLexeme(3);
  if (!label)
    return -1;

  // Set up defaults
  Bool multiple = FALSE;
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;

  if (no_args > 3)
    multiple = (Bool)RtnLong(4);

  if (no_args > 4)
    x = (int) RtnLong(5);

  if (no_args > 5)
    y = (int) RtnLong(6);

  if (no_args > 6)
    width = (int) RtnLong(7);

  if (no_args > 7)
    height = (int) RtnLong(8);

  wxPanel *panel = (wxPanel *)wxGetTypedObject(parent_id, wxTYPE_PANEL);
  if (panel)
  {
    wxClipsListBox *item = new wxClipsListBox(panel, func, (wxFunction)wxClipsListBoxCallback, label, multiple, x, y, width, height);
    wxStoreThingInTable((long)item, item, wxTYPE_LIST_BOX);
    return (long)item;
  }
  else
  {
    ClipsError << "Error in list-box-create: could not find parent panel " << parent_id << ".\n";
    return 0;
  }
}

// Append a string (and optional client data)
long clipsListBoxAppend()
{
  if (ArgCountCheck("clipsListBoxAppend", AT_LEAST, 2) == -1)
    return 0;

  int no_args = RtnArgCount();

  long id = RtnLong(1);
  char *item_string = RtnLexeme(2);
  if (!item_string)
    return 0;

  char *client_data = NULL;
  if (no_args > 2)
    client_data = RtnLexeme(3);

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    item->Append(item_string, client_data);
    return 1;
  }
  else
  {
    ClipsError << "Error in list-box-append: could not find list box " << id << ".\n";
    return 0;
  }
}

// Find a string and return ID
long clipsListBoxFindString()
{
  if (ArgCountCheck("clipsListBoxFindString", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *item_string = RtnLexeme(2);
  if (!item_string)
    return -1;

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    return item->FindString(item_string);
  }
  else
  {
    ClipsError << "Error in list-box-find-string: could not find list box " << id << ".\n";
    return -1;
  }
}

// Clear
long clipsListBoxClear()
{
  if (ArgCountCheck("clipsListBoxClear", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    item->Clear();
    return 1;
  }
  else
  {
    ClipsError << "Error in list-box-clear: could not find list box " << id << ".\n";
    return 0;
  }
}

// Set the current selection
long clipsListBoxSetSelection()
{
  if (ArgCountCheck("clipsListBoxSetSelection", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  int sel = (int)RtnLong(2);

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    item->SetSelection(sel);
    return 1;
  }
  else
  {
    ClipsError << "Error in list-box-set-selection: could not find list box " << id << ".\n";
    return 0;
  }
}

// Deselect an item
long clipsListBoxDeselect()
{
  if (ArgCountCheck("clipsListBoxDeselect", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  int sel = (int)RtnLong(2);

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    item->Deselect(sel);
    return 1;
  }
  else
  {
    ClipsError << "Error in list-box-deselect: could not find list box " << id << ".\n";
    return 0;
  }
}

// Get the current selection
long clipsListBoxGetSelection()
{
  if (ArgCountCheck("clipsListBoxGetSelection", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    return item->GetSelection();
  }
  else
  {
    ClipsError << "Error in list-box-get-selection: could not find list box " << id << ".\n";
    return -1;
  }
}

// Get string selection
char *clipsListBoxGetStringSelection()
{
  if (ArgCountCheck("clipsListBoxGetStringSelection", EXACTLY, 1) == -1)
    return (char *)AddSymbol("");

  long id = RtnLong(1);

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    char *s = item->GetStringSelection();
    if (s)
      return (char *)AddSymbol(s);
    else
      return (char *)AddSymbol("");
  }
  else
  {
    ClipsError << "Error in list-box-get-string-selection: could not find list box " << id << ".\n";
    return (char *)AddSymbol("");
  }
}

// Set string selection
long clipsListBoxSetStringSelection()
{
  if (ArgCountCheck("clipsListBoxSetStringSelection", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *sel = RtnLexeme(2);
  if (!sel)
    return 0;

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    return item->SetStringSelection(sel);
  }
  else
  {
    ClipsError << "Error in list-box-set-string-selection: could not find list box " << id << ".\n";
    return 0;
  }
}

// Get number of items
long clipsListBoxNumber()
{
  if (ArgCountCheck("clipsListBoxSetNumber", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    return item->Number();
  }
  else
  {
    ClipsError << "Error in list-box-number: could not find list box " << id << ".\n";
    return -1;
  }
}

// Delete one item
long clipsListBoxDelete()
{
  if (ArgCountCheck("clipsListBoxSetNumber", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  int n = (int) RtnLong(2);

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    item->Delete(n);
    return 1;
  }
  else
  {
    ClipsError << "Error in list-box-delete: could not find list box " << id << ".\n";
    return 0;
  }
}


// Get client data for an item
char *clipsListBoxGetClientData()
{
  if (ArgCountCheck("clipsListBoxGetClientData", EXACTLY, 2) == -1)
    return (char *)AddSymbol("");

  long id = RtnLong(1);
  int sel = (int)RtnLong(2);

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    char *s = item->GetClientData(sel);
    if (s)
      return (char *)AddSymbol(s);
    else
      return (char *)AddSymbol("");
  }
  else
  {
    ClipsError << "Error in list-box-get-client-data: could not find list box " << id << ".\n";
    return (char *)AddSymbol("");
  }
}

// Get string for integer ID
char *clipsListBoxGetString()
{
  if (ArgCountCheck("clipsListBoxGetString", EXACTLY, 2) == -1)
    return (char *)AddSymbol("");

  long id = RtnLong(1);
  int sel = (int)RtnLong(2);

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    char *s = item->GetString(sel);
    if (s)
      return (char *)AddSymbol(s);
    else
      return (char *)AddSymbol("");
  }
  else
  {
    ClipsError << "Error in list-box-get-string: could not find list box " << id << ".\n";
    return (char *)AddSymbol("");
  }
}

// Get selections
int *wxListBoxSelections = NULL;
int wxListBoxNSelections = 0;
int wxListBoxCurrentSelection;

long clipsListBoxGetFirstSelection()
{
  if (ArgCountCheck("clipsListBoxGetFirstSelection", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxListBox *item = (wxListBox *)wxGetTypedObject(id, wxTYPE_LIST_BOX);
  if (item)
  {
    wxListBoxNSelections = item->GetSelections(&wxListBoxSelections);
    wxListBoxCurrentSelection = 1;
    if (wxListBoxNSelections > 0)
      return wxListBoxSelections[0];
    else
      return -1;
  }
  else
  {
    ClipsError << "Error in list-box-get-first-selection: could not find list box " << id << ".\n";
    return -1;
  }
}

long clipsListBoxGetNextSelection()
{
  if (ArgCountCheck("clipsListBoxGetNextSelection", EXACTLY, 0) == -1)
    return 0;

  if (wxListBoxCurrentSelection >= wxListBoxNSelections)
    return -1;

  wxListBoxCurrentSelection ++;

  return wxListBoxSelections[wxListBoxCurrentSelection - 1];
}

// Create a choice widget
long clipsChoiceCreate()
{
  if (ArgCountCheck("clipsChoiceCreate", AT_LEAST, 3) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *func = RtnLexeme(2);
  if (func && (strlen(func) == 0))
    func = NULL;

  char *label = RtnLexeme(3);
  if (!label)
    return -1;

  // Set up defaults
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;

  if (no_args > 3)
    x = (int) RtnLong(4);

  if (no_args > 4)
    y = (int) RtnLong(5);

  if (no_args > 5)
    width = (int) RtnLong(6);

  if (no_args > 6)
    height = (int) RtnLong(7);

  wxPanel *panel = (wxPanel *)wxGetTypedObject(parent_id, wxTYPE_PANEL);
  if (panel)
  {
    wxClipsChoice *item = new wxClipsChoice(panel, func, (wxFunction)wxClipsChoiceCallback, label, x, y, width, height);
    wxStoreThingInTable((long)item, item, wxTYPE_CHOICE);
    return (long)item;
  }
  else
  {
    ClipsError << "Error in choice-create: could not find parent panel " << parent_id << ".\n";
    return 0;
  }
}

// Append a string
long clipsChoiceAppend()
{
  if (ArgCountCheck("clipsChoiceAppend", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *item_string = RtnLexeme(2);
  if (!item_string)
    return 0;

  wxChoice *item = (wxChoice *)wxGetTypedObject(id, wxTYPE_CHOICE);
  if (item)
  {
    item->Append(item_string);
    return 1;
  }
  else
  {
    ClipsError << "Error in choice-append: could not find choice " << id << ".\n";
    return 0;
  }
}

// Find a string and return ID
long clipsChoiceFindString()
{
  if (ArgCountCheck("clipsChoiceFindString", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *item_string = RtnLexeme(2);
  if (!item_string)
    return -1;

  wxChoice *item = (wxChoice *)wxGetTypedObject(id, wxTYPE_CHOICE);
  if (item)
  {
    return item->FindString(item_string);
  }
  else
  {
    ClipsError << "Error in choice-find-string: could not find choice " << id << ".\n";
    return -1;
  }
}

// Clear
long clipsChoiceClear()
{
  if (ArgCountCheck("clipsChoiceClear", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxChoice *item = (wxChoice *)wxGetTypedObject(id, wxTYPE_CHOICE);
  if (item)
  {
    item->Clear();
    return 1;
  }
  else
  {
    ClipsError << "Error in choice-clear: could not find choice " << id << ".\n";
    return 0;
  }
}

// Set the current selection
long clipsChoiceSetSelection()
{
  if (ArgCountCheck("clipsChoiceSetSelection", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  int sel = (int)RtnLong(2);

  wxChoice *item = (wxChoice *)wxGetTypedObject(id, wxTYPE_CHOICE);
  if (item)
  {
    item->SetSelection(sel);
    return 1;
  }
  else
  {
    ClipsError << "Error in choice-set-selection: could not find choice " << id << ".\n";
    return 0;
  }
}

// Get the current selection
long clipsChoiceGetSelection()
{
  if (ArgCountCheck("clipsChoiceGetSelection", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxChoice *item = (wxChoice *)wxGetTypedObject(id, wxTYPE_CHOICE);
  if (item)
  {
    return item->GetSelection();
  }
  else
  {
    ClipsError << "Error in choice-get-selection: could not find choice " << id << ".\n";
    return -1;
  }
}

// Get string selection
char *clipsChoiceGetStringSelection()
{
  if (ArgCountCheck("clipsChoiceGetStringSelection", EXACTLY, 1) == -1)
    return (char *)AddSymbol("");

  long id = RtnLong(1);

  wxChoice *item = (wxChoice *)wxGetTypedObject(id, wxTYPE_CHOICE);
  if (item)
  {
    char *s = item->GetStringSelection();
    if (s)
      return (char *)AddSymbol(s);
    else
      return (char *)AddSymbol("");
  }
  else
  {
    ClipsError << "Error in choice-get-string-selection: could not find choice " << id << ".\n";
    return (char *)AddSymbol("");
  }
}

// Set string selection
long clipsChoiceSetStringSelection()
{
  if (ArgCountCheck("clipsChoiceSetStringSelection", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *sel = RtnLexeme(2);
  if (!sel)
    return 0;

  wxChoice *item = (wxChoice *)wxGetTypedObject(id, wxTYPE_CHOICE);
  if (item)
  {
    return item->SetStringSelection(sel);
  }
  else
  {
    ClipsError << "Error in choice-set-string-selection: could not find choice " << id << ".\n";
    return 0;
  }
}

// Get string for integer ID
char *clipsChoiceGetString()
{
  if (ArgCountCheck("clipsChoiceGetString", EXACTLY, 2) == -1)
    return (char *)AddSymbol("");

  long id = RtnLong(1);
  int sel = (int)RtnLong(2);

  wxChoice *item = (wxChoice *)wxGetTypedObject(id, wxTYPE_CHOICE);
  if (item)
  {
    char *s = item->GetString(sel);
    if (s)
      return (char *)AddSymbol(s);
    else
      return (char *)AddSymbol("");
  }
  else
  {
    ClipsError << "Error in choice-get-string: could not find choice " << id << ".\n";
    return (char *)AddSymbol("");
  }
}


// Create a menu
long clipsMenuCreate()
{
  if (ArgCountCheck("clipsMenuCreate", AT_LEAST, 0) == -1)
    return 0;

  int no_args = RtnArgCount();

  // Set up defaults
  char *title = NULL;
  if (no_args > 0)
    title = RtnLexeme(1);

  wxClipsMenu *menu = new wxClipsMenu(title);
  wxStoreThingInTable((long)menu, menu, wxTYPE_MENU);
  return (long)menu;
}

// Append a string or submenu
long clipsMenuAppend()
{
  if (ArgCountCheck("clipsMenuAppend", AT_LEAST, 3) == -1)
    return 0;

  int no_args = RtnArgCount();

  long id = RtnLong(1);
  int item_id = (int)RtnLong(2);
  char *menu_string = RtnLexeme(3);
  if (!menu_string)
    return 0;

  wxMenu *menu = (wxMenu *)wxGetTypedObject(id, wxTYPE_MENU);
  if (menu)
  {
    if (no_args > 3)
    {
      long submenu_id = RtnLong(4);
      wxMenu *submenu = (wxMenu *)wxGetTypedObject(submenu_id, wxTYPE_MENU);
      if (submenu)
      {
        menu->Append(item_id, menu_string, submenu);
        return 1;
      }
      else
      {
        ClipsError << "Error in menu-append: could not find submenu " << submenu_id << "\n";
        return 0;
      }
    }
    else
    {
      menu->Append(item_id, menu_string);
      return 1;
    }
  }
  else
  {
    ClipsError << "Error in menu-append: could not find menu " << id << ".\n";
    return 0;
  }
}

// Append a separator
long clipsMenuAppendSeparator()
{
  if (ArgCountCheck("clipsMenuAppendSeparator", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxMenu *menu = (wxMenu *)wxGetTypedObject(id, wxTYPE_MENU);
  if (menu)
  {
    menu->AppendSeparator();
    return 1;
  }
  else
  {
    ClipsError << "Error in menu-append-separator: could not find menu " << id << ".\n";
    return 0;
  }
}

// Enable an item
long clipsMenuEnable()
{
  if (ArgCountCheck("clipsMenuEnable", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  int item_id = (int)RtnLong(2);
  Bool enabled = (Bool)RtnLong(3);

  wxMenu *menu = (wxMenu *)wxGetTypedObject(id, wxTYPE_MENU);
  if (menu)
  {
    menu->Enable(item_id, enabled);
    return 1;
  }
  else
  {
    ClipsError << "Error in menu-enable: could not find menu " << id << ".\n";
    return 0;
  }
}

// Check an item
long clipsMenuCheck()
{
  if (ArgCountCheck("clipsMenuCheck", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  int item_id = (int)RtnLong(2);
  Bool checked = (Bool)RtnLong(3);

  wxMenu *menu = (wxMenu *)wxGetTypedObject(id, wxTYPE_MENU);
  if (menu)
  {
    menu->Check(item_id, checked);
    return 1;
  }
  else
  {
    ClipsError << "Error in menu-check: could not find menu " << id << ".\n";
    return 0;
  }
}

// Create a menu bar
long clipsMenuBarCreate()
{
  if (ArgCountCheck("clipsMenuBarCreate", EXACTLY, 0) == -1)
    return 0;

  wxClipsMenuBar *menu_bar = new wxClipsMenuBar;
  wxStoreThingInTable((long)menu_bar, menu_bar, wxTYPE_MENU_BAR);
  return (long)menu_bar;
}

// Append a menu
long clipsMenuBarAppend()
{
  if (ArgCountCheck("clipsMenuBarAppend", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  long menu_id = RtnLong(2);
  char *menu_string = RtnLexeme(3);
  if (!menu_string)
    return 0;

  wxMenuBar *menu_bar = (wxMenuBar *)wxGetTypedObject(id, wxTYPE_MENU_BAR);
  if (menu_bar)
  {
    wxMenu *menu = (wxMenu *)wxGetTypedObject(menu_id, wxTYPE_MENU);
    if (menu)
    {
      menu_bar->Append(menu, menu_string);
      return 1;
    }
    else
    {
      ClipsError << "Error in menu-bar-append: could not find menu " << menu_id << ".\n";
      return 0;
    }
  }
  else
  {
    ClipsError << "Error in menu-bar-append: could not find menu bar " << id << ".\n";
    return 0;
  }
}

// Enable an item
long clipsMenuBarEnable()
{
  if (ArgCountCheck("clipsMenuBarEnable", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  int item_id = (int)RtnLong(2);
  Bool enabled = (Bool)RtnLong(3);

  wxMenuBar *menu_bar = (wxMenuBar *)wxGetTypedObject(id, wxTYPE_MENU_BAR);
  if (menu_bar)
  {
    menu_bar->Enable(item_id, enabled);
    return 1;
  }
  else
  {
    ClipsError << "Error in menu-bar-enable: could not find menu bar " << id << ".\n";
    return 0;
  }
}

// Check an item
long clipsMenuBarCheck()
{
  if (ArgCountCheck("clipsMenuBarCheck", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  int item_id = (int)RtnLong(2);
  Bool checked = (Bool)RtnLong(3);

  wxMenuBar *menu_bar = (wxMenuBar *)wxGetTypedObject(id, wxTYPE_MENU_BAR);
  if (menu_bar)
  {
    menu_bar->Check(item_id, checked);
    return 1;
  }
  else
  {
    ClipsError << "Error in menu-bar-check: could not find menu bar " << id << ".\n";
    return 0;
  }
}


/************************************************************************
 * FRAME FUNCTIONS
 *
 ************************************************************************/

// Create a new dialog box
long clipsFrameCreate()
{
  if (ArgCountCheck("clipsFrameCreate", AT_LEAST, 2) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *title = RtnLexeme(2);
  if (!title)
    return -1;

  // Set up defaults

  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;
  int style = wxSDI | wxDEFAULT_FRAME;

  if (no_args > 2)
    x = (int) RtnLong(3);

  if (no_args > 3)
    y = (int) RtnLong(4);

  if (no_args > 4)
    width = (int) RtnLong(5);

  if (no_args > 5)
    height = (int) RtnLong(6);

  if (no_args > 6)
  {
    char *s = RtnLexeme(7);
    if (!s)
      return -1;
    if (strcmp(s, "wxMDI_PARENT") == 0)
      style = wxMDI_PARENT;
    else if (strcmp(s, "wxSDI") == 0)
      style = wxSDI;
    else if (strcmp(s, "wxMDI_CHILD") == 0)
      style = wxMDI_CHILD;
  }

  wxFrame *frame = NULL;

  if (parent_id != 0)
  {
    frame = (wxFrame *)wxGetTypedObject(parent_id, wxTYPE_FRAME);
    if (!frame)
    {
      ClipsError << "Error in frame-create: could not find parent frame " << parent_id << ".\n";
      return 0;
    }
  }

  wxClipsFrame *new_frame = new wxClipsFrame(frame, title, x, y, width, height, style);
  wxStoreThingInTable((long)new_frame, new_frame, wxTYPE_FRAME);
  return (long)new_frame;
}

// Set the menu bar
long clipsFrameSetMenuBar()
{
  if (ArgCountCheck("clipsFrameSetMenuBar", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long menubar_id = RtnLong(2);

  wxFrame *frame = (wxFrame *)wxGetTypedObject(id, wxTYPE_FRAME);
  if (frame)
  {
    wxMenuBar *menu_bar = (wxMenuBar *)wxGetTypedObject(menubar_id, wxTYPE_MENU_BAR);
    if (menu_bar)
    {
      frame->SetMenuBar(menu_bar);
      return 1;
    }
    else
    {
      ClipsError << "Error in frame-set-menu-bar: could not find menu bar " << menubar_id << ".\n";
      return 0;
    }
  }
  else
  {
    ClipsError << "Error in frame-set-menu-bar: could not find frame " << id << ".\n";
    return 0;
  }
}

// Create the frame's status line
long clipsFrameCreateStatusLine()
{
  if (ArgCountCheck("clipsFrameCreateStatusLine", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  wxFrame *frame = (wxFrame *)wxGetTypedObject(id, wxTYPE_FRAME);
  if (frame)
  {
    frame->CreateStatusLine();
    return 1;
  }
  else
  {
    ClipsError << "Error in frame-create-status-line: could not find frame " << id << ".\n";
    return 0;
  }
}

// Set the status line text
long clipsFrameSetStatusText()
{
  if (ArgCountCheck("clipsFrameSetStatusText", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *text = RtnLexeme(2);
  if (!text)
    return 0;

  wxFrame *frame = (wxFrame *)wxGetTypedObject(id, wxTYPE_FRAME);
  if (frame)
  {
    frame->SetStatusText(text);
    return 1;
  }
  else
  {
    ClipsError << "Error in frame-set-status-text: could not find frame " << id << ".\n";
    return 0;
  }
}

// Set the icon
long clipsFrameSetIcon()
{
  if (ArgCountCheck("clipsFrameSetIcon", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long iconId = RtnLong(2);

  wxFrame *frame = (wxFrame *)wxGetTypedObject(id, wxTYPE_FRAME);
  if (frame)
  {
    wxIcon *icon = (wxIcon *)wxGetTypedObject(iconId, wxTYPE_ICON);
    if (icon)
    {
      frame->SetIcon(icon);
      return 1;
    }
    else
    {
      ClipsError << "Error in frame-set-icon: could not find icon " << iconId << ".\n";
      return 0;
    }
  }
  else
  {
    ClipsError << "Error in frame-set-icon: could not find frame " << id << ".\n";
    return 0;
  }
}

// Iconimize/restore the frame
long clipsFrameIconize()
{
  if (ArgCountCheck("clipsFrameIconize", AT_LEAST, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long iconize = TRUE;

  if (RtnArgCount() > 1)
    iconize = RtnLong(2);

  wxFrame *frame = (wxFrame *)wxGetTypedObject(id, wxTYPE_FRAME);
  if (frame)
  {
    frame->Iconize((Bool)iconize);
    return 1;
  }
  else
  {
    ClipsError << "Error in frame-iconize: could not find frame " << id << ".\n";
    return 0;
  }
}


/************************************************************************
 * FUNCTIONS TO ACCESS HELP SYSTEM
 *
 ************************************************************************/

/*
extern wxHelpInstance *HelpInstance;

long clipsHelpLoadFile()
{
  if (ArgCountCheck("clipsHelpLoadFile", EXACTLY, 1) == -1)
    return 0;
  char *filename = RtnLexeme(1);
  if (!filename)
    return 0;
  return HelpInstance->LoadFile(filename);
}

long clipsHelpDisplayContents()
{
  if (ArgCountCheck("clipsHelpDisplayContents", EXACTLY, 0) == -1)
    return 0;
  return HelpInstance->DisplayContents();
}

long clipsHelpDisplaySection()
{
  if (ArgCountCheck("clipsHelpDisplaySection", EXACTLY, 1) == -1)
    return 0;
  return HelpInstance->DisplaySection(RtnLong(1));
}

long clipsHelpDisplayBlock()
{
  if (ArgCountCheck("clipsHelpDisplayBlock", EXACTLY, 1) == -1)
    return 0;
  return HelpInstance->DisplayBlock(RtnLong(1));
}

long clipsHelpKeywordSearch()
{
  if (ArgCountCheck("clipsHelpKeywordSearch", EXACTLY, 1) == -1)
    return 0;
  char *keyword = RtnLexeme(1);
  if (!keyword)
    return 0;
  return HelpInstance->KeywordSearch(keyword);
}

*/

/************************************************************************
 * INPUT/OUTPUT AND USER INTERFACE FUNCTIONS
 *
 ************************************************************************/

/*
 * Implement reading strings and symbols
 *
 */

/*
char *clipsRead(void)
{
  char *s = wxGetTextFromUser("Enter text");
  if (s)
    return (char *)AddSymbol(s);
  return (char *)AddSymbol("");
}
*/

char *clipsReadString(void)
{
  char *s = wxGetTextFromUser("Enter text");
  if (s)
    return (char *)AddSymbol(s);
  return (char *)AddSymbol("");
}

/*
 * Pass optional message and optional default string
 *
 */

char *clipsGetTextFromUser(void)
{
  char *message = "Enter a string";
  char *default_string = "";
  Bool centre = TRUE;

  if (ArgCountCheck("clipsGetTextFromUser", AT_LEAST, 1) != -1)
    message = RtnLexeme(1);

  if (!message)
    return (char *)AddSymbol("");

  int no_args = RtnArgCount();

  if (no_args > 1)
    default_string = RtnLexeme(2);
  if (!default_string)
    return (char *)AddSymbol("");

  if (no_args > 2)
    centre = (Bool)RtnLong(3);

  char *s = wxGetTextFromUser(message, "Enter text", default_string, NULL, -1, -1, centre);
  if (s)
    return (char *)AddSymbol(s);
  return (char *)AddSymbol("");
}

/*
 * Convert string to number
 *
 */

double clipsStringToFloat(void)
{
  if (ArgCountCheck("clipsStringToFloat", EXACTLY, 1) == -1)
  {
    ClipsError << "Error in string-to-float: one string argument required.\n";
    return 0.0;
  }

  char *str = RtnLexeme(1);
  if (!str)
    return 0.0;
  return atof(str);
}

/*
 * Convert number to string
 *
 */

char *clipsFloatToString(void)
{
  if (ArgCountCheck("clipsFloatToString", EXACTLY, 1) == -1)
  {
    ClipsError << "Error in float-to-string: one floating point argument required.\n";
    return (char *)AddSymbol("");
  }
  char buf[50];
  sprintf(buf, "%.4f", RtnDouble(1));
  return (char *)AddSymbol(buf);
}

/*
 * Convert a list of strings or symbols to a sorted list of strings
 *
 */

VOID clipsStringSort(DATA_OBJECT_PTR returnValuePtr)
{
  DATA_OBJECT arg;

  if (ArgCountCheck("clipsStringSort", AT_LEAST, 1) == -1)
  {
    ClipsError << "Error in string-sort: 1st argument must be a multifield of strings or symbols.\n";
    SetMultifieldErrorValue(returnValuePtr);
    return;
  }

  RtnUnknown(1, &arg);

  if (GetType(arg) != MULTIFIELD)
  {
    ClipsError << "Error in string-sort: 1st argument must be a multifield of strings or symbols.\n";
    SetMultifieldErrorValue(returnValuePtr);
    return;
  }

  int start = GetpDOBegin(&arg);
  int end = GetpDOEnd(&arg);
  int no_strings = end-start + 1;

  VOID *multifieldPtr = ClipsGetValue(arg);
  wxStringList stringList;

  int i;
  for (i = start; i <= end; i++)
  {
    if ((GetMFType(multifieldPtr, i) == STRING) || (GetMFType(multifieldPtr, i) == STRING))
      stringList.Add(ValueToString(GetMFValue(multifieldPtr, i)));
    else
    {
      ClipsError << "Error in string-sort: multifield element " << i-start+1 << " must be a string or symbol.\n";
      SetMultifieldErrorValue(returnValuePtr);
      return;
    }
  }
  stringList.Sort();

  VOID *newMultifield = CreateMultifield(no_strings);
  for (i = start; i <= end; i++)
  {
    wxNode *node = stringList.Nth(i-start);
    if (node)
    {
      char *s = (char *)node->Data();

      SetMFType(newMultifield, i, STRING);
      SetMFValue(newMultifield, i, AddSymbol(s));
    }
  }
  SetpType(returnValuePtr, MULTIFIELD);
  SetpValue(returnValuePtr, newMultifield);
  SetpDOBegin(returnValuePtr,start);
  SetpDOEnd(returnValuePtr,end);
  return;
}


/*
 * Ask user to choose from a menu of strings (message followed by a
 * variable number of string arguments)
 *
 */

char *clipsGetChoice(void)
{
  DATA_OBJECT arg;
  Bool centre = TRUE;

  if (ArgCountCheck("clipsGetChoice", AT_LEAST, 2) == -1)
    return (char *)AddSymbol("");

  int no_args = RtnArgCount();
  if (no_args > 2)
    centre = (Bool)RtnLong(3);

  RtnUnknown(2, &arg);

  if (GetType(arg) != MULTIFIELD)
  {
    ClipsError << "Error in get-choice: 2nd argument must be a multifield of strings.\n";
    return (char *)AddSymbol("");
  }

  char *message = RtnLexeme(1);
  if (!message)
    return (char *)AddSymbol("");

  int start = GetpDOBegin(&arg);
  int end = GetpDOEnd(&arg);
  int no_strings = end-start + 1;

  VOID *multifieldPtr = ClipsGetValue(arg);

  char **choices = new char *[no_strings];

  for (int i = start; i <= end; i++)
  {
    if ((GetMFType(multifieldPtr, i) == STRING) || (GetMFType(multifieldPtr, i) == STRING))
      choices[i-start] = ValueToString(GetMFValue(multifieldPtr, i));
    else
    {
      ClipsError << "Error in get-choice: multifield element " << i-start+1 << " must be a string or symbol.\n";
      delete choices;
      return (char *)AddSymbol("");
    }
  }

  char *s = wxGetSingleChoice(message, "Menu", no_strings, choices, NULL, -1, -1, centre);
  delete choices;

  if (s)
    return (char *)AddSymbol(s);
  return (char *)AddSymbol("");
}

/*
 * Pass message, optional flags = "OK", "OK-CANCEL", "YES-NO" or "YES-NO-CANCEL"
 *
 */

char *clipsMessageBox(void)
{
  int no_args = RtnArgCount();
  if (ArgCountCheck("clipsMessageBox", AT_LEAST, 1) == -1)
    return (char *)AddSymbol("");

  char *message = RtnLexeme(1);
  if (!message)
    return (char *)AddSymbol("");

  char *flags = "OK";
  Bool centre = TRUE;

  if (no_args > 1)
    flags = RtnLexeme(2);
  if (!flags)
    return (char *)AddSymbol("");

  if (no_args > 2)
    centre = (Bool)RtnLong(3);

  int wx_flags = wxOK;
  if (strcmp(flags, "OK") == 0)
    wx_flags = wxOK;
  else if (strcmp(flags, "OK-CANCEL") == 0)
    wx_flags = wxOK | wxCANCEL;
  else if (strcmp(flags, "YES-NO") == 0)
    wx_flags = wxYES_NO;
  else if (strcmp(flags, "YES-NO-CANCEL") == 0)
    wx_flags = wxYES_NO | wxCANCEL;

  int res = wxMessageBox(message, "Message", wx_flags, NULL, -1, -1, centre);
  char *ans = "ERROR";
  switch (res)
  {  
    case wxCANCEL: ans = "CANCEL"; break;
    case wxOK:     ans = "OK"; break;
    case wxYES:    ans = "YES"; break;
    case wxNO:     ans = "NO"; break;
  }

  return (char *)AddSymbol(ans);
}

/*
 * Pass optional message, default path, default file, default extension, default
 * wildcard.
 *
 */

char *clipsFileSelector(void)
{
  char *message = "Select a file";
  char *def_path = NULL;
  char *def_file = NULL;
  char *def_ext = NULL;
  char *def_wildcard = "*.*";

  int no_args = RtnArgCount();
  if (no_args > 0)
    message = RtnLexeme(1);
  if (!message)
    return (char *)AddSymbol("");

  if (no_args > 1)
    def_path = RtnLexeme(2);
  if (no_args > 2)
    def_file = RtnLexeme(3);
  if (no_args > 3)
    def_ext = RtnLexeme(4);
  if (no_args > 4)
    def_wildcard = RtnLexeme(5);
  if (!def_wildcard)
    return (char *)AddSymbol("");

  char *s = wxFileSelector(message, def_path, def_file, def_ext, def_wildcard);

  if (s)
    return (char *)AddSymbol(s);
  return (char *)AddSymbol("");
}


/* Have to implement new batch command since CLIPS assumes using
 * the main loop
 */
void wxExecuteClipsFile(char *filename)
{
  if (FileExists(filename))
  {
    ExecuteBatchFile(filename);
  }
#ifdef wx_x
  else cerr << "Cannot find CLIPS file " << filename << ".\n";
#endif
}

void clipsBatch(void)
{
  if (ArgCountCheck("clipsBatch", EXACTLY, 1) == -1)
    return;
  char *filename = RtnLexeme(1);
  if (!filename)
    return;
  wxExecuteClipsFile(filename);
}


/*************************************************************/
/* UserFunctions:  The function which informs CLIPS of any   */
/*   user defined functions.  In the default case, there are */
/*   no user defined functions.  To define functions, either */
/*   this function must be replaced by a function with the   */
/*   same name within this file, or this function can be     */
/*   deleted from this file and included in another file.    */
/*   User defined functions may be included in this file or  */
/*   other files.                                            */
/*   Example of redefined UserFunctions:                     */
/*     UserFunctions()                                       */
/*       {                                                   */
/*        DefineFunction("fun1",'i',fun1,"fun1");            */
/*        DefineFunction("other",'f',other,"other");         */
/*       }                                                   */
/*************************************************************/

void wxUserFunctions()
{
  // Miscellaneous functions
  DefineFunction("batch", 'v', PTIF clipsBatch, "clipsBatch");
  DefineFunction("file-selector", 's', PTIF clipsFileSelector, "clipsFileSelector");
  DefineFunction("float-to-string", 's', PTIF clipsFloatToString, "clipsFloatToString");
  DefineFunction("get-choice", 's', PTIF clipsGetChoice, "clipsGetChoice");
  DefineFunction("get-text-from-user", 's', PTIF clipsGetTextFromUser, "clipsGetTextFromUser");
  DefineFunction("get-platform", 's', PTIF clipsGetPlatform, "clipsGetPlatform");
  DefineFunction("message-box", 'w', PTIF clipsMessageBox, "clipsMessageBox");
  DefineFunction("read-string", 's', PTIF clipsReadString, "clipsReadString");
  DefineFunction("string-to-float", 'd', PTIF clipsStringToFloat, "clipsStringToFloat");

  // Window functions
  DefineFunction("brush-create", 'l', PTIF clipsBrushCreate, "clipsBrushCreate");
  DefineFunction("brush-delete", 'l', PTIF clipsBrushDelete, "clipsBrushDelete");

  DefineFunction("button-create", 'l', PTIF clipsButtonCreate, "clipsButtonCreate");

  DefineFunction("canvas-create", 'l', PTIF clipsCanvasCreate, "clipsCanvasCreate");
  DefineFunction("canvas-set-scrollbars", 'l', PTIF clipsCanvasSetScrollbars, "clipsCanvasSetScrollbars");
  DefineFunction("canvas-scroll", 'l', PTIF clipsCanvasScroll, "clipsCanvasScroll");
  DefineFunction("canvas-get-dc", 'l', PTIF clipsCanvasGetDC, "clipsCanvasGetDC");

  DefineFunction("check-box-create", 'l', PTIF clipsCheckBoxCreate, "clipsCheckBoxCreate");
  DefineFunction("check-box-set-value", 'l', PTIF clipsCheckBoxSetValue, "clipsCheckBoxSetValue");
  DefineFunction("check-box-get-value", 'l', PTIF clipsCheckBoxGetValue, "clipsCheckBoxGetValue");

  DefineFunction("choice-create", 'l', PTIF clipsChoiceCreate, "clipsChoiceCreate");
  DefineFunction("choice-append", 'l', PTIF clipsChoiceAppend, "clipsChoiceAppend");
  DefineFunction("choice-find-string", 'l', PTIF clipsChoiceFindString, "clipsChoiceFindString");
  DefineFunction("choice-clear", 'l', PTIF clipsChoiceClear, "clipsChoiceClear");
  DefineFunction("choice-set-selection", 'l', PTIF clipsChoiceSetSelection, "clipsChoiceSetSelection");
  DefineFunction("choice-get-string-selection", 's', PTIF clipsChoiceGetStringSelection, "clipsChoiceGetStringSelection");
  DefineFunction("choice-get-selection", 'l', PTIF clipsChoiceGetSelection, "clipsChoiceGetSelection");
  DefineFunction("choice-set-string-selection", 'l', PTIF clipsChoiceSetStringSelection, "clipsChoiceSetStringSelection");
  DefineFunction("choice-get-string", 's', PTIF clipsChoiceGetString, "clipsChoiceGetString");

  DefineFunction("clean-windows", 'v', PTIF clipsCleanWindows, "clipsCleanWindows");

  DefineFunction("dialog-box-create", 'l', PTIF clipsDialogBoxCreate, "clipsDialogBoxCreate");

  DefineFunction("dc-clear", 'l', PTIF clipsDCClear, "clipsDCClear");
  DefineFunction("dc-destroy-clipping-region", 'l', PTIF clipsDCDestroyClippingRegion, "clipsDCDestroyClippingRegion");
  DefineFunction("dc-draw-ellipse", 'l', PTIF clipsDCDrawEllipse, "clipsDCDrawEllipse");
  DefineFunction("dc-draw-line", 'l', PTIF clipsDCDrawLine, "clipsDCDrawLine");
  DefineFunction("dc-draw-lines", 'l', PTIF clipsDCDrawLines, "clipsDCDrawLines");
  DefineFunction("dc-draw-point", 'l', PTIF clipsDCDrawPoint, "clipsDCDrawPoint");
  DefineFunction("dc-draw-polygon", 'l', PTIF clipsDCDrawPolygon, "clipsDCDrawPolygon");
  DefineFunction("dc-draw-spline", 'l', PTIF clipsDCDrawSpline, "clipsDCDrawSpline");
  DefineFunction("dc-draw-rectangle", 'l', PTIF clipsDCDrawRectangle, "clipsDCDrawRectangle");
  DefineFunction("dc-draw-rounded-rectangle", 'l', PTIF clipsDCDrawRoundedRectangle, "clipsDCDrawRoundedRectangle");
  DefineFunction("dc-draw-text", 'l', PTIF clipsDCDrawText, "clipsDCDrawText");
  DefineFunction("dc-set-clipping-region", 'l', PTIF clipsDCSetClippingRegion, "clipsDCSetClippingRegion");
  DefineFunction("dc-set-pen", 'l', PTIF clipsDCSetPen, "clipsDCSetPen");
  DefineFunction("dc-set-logical-function", 'l', PTIF clipsDCSetLogicalFunction, "clipsDCSetLogicalFunction");
  DefineFunction("dc-set-text-foreground", 'l', PTIF clipsDCSetTextForeground, "clipsDCSetTextForeground");
  DefineFunction("dc-set-text-background", 'l', PTIF clipsDCSetTextBackground, "clipsDCSetTextBackground");
  DefineFunction("dc-set-brush", 'l', PTIF clipsDCSetBrush, "clipsDCSetBrush");
  DefineFunction("dc-set-font", 'l', PTIF clipsDCSetFont, "clipsDCSetFont");

  DefineFunction("mouse-event-is-button", 'l', PTIF clipsMouseEventIsButton, "clipsMouseEventIsButton");
  DefineFunction("mouse-event-button-down", 'l', PTIF clipsMouseEventButtonDown, "clipsMouseEventButtonDown");
  DefineFunction("mouse-event-control-down", 'l', PTIF clipsMouseEventControlDown, "clipsMouseEventControlDown");
  DefineFunction("mouse-event-shift-down", 'l', PTIF clipsMouseEventShiftDown, "clipsMouseEventShiftDown");
  DefineFunction("mouse-event-button", 'l', PTIF clipsMouseEventButton, "clipsMouseEventButton");
  DefineFunction("mouse-event-left-down", 'l', PTIF clipsMouseEventLeftDown, "clipsMouseEventLeftDown");
  DefineFunction("mouse-event-middle-down", 'l', PTIF clipsMouseEventMiddleDown, "clipsMouseEventMiddleDown");
  DefineFunction("mouse-event-right-down", 'l', PTIF clipsMouseEventRightDown, "clipsMouseEventRightDown");
  DefineFunction("mouse-event-left-up", 'l', PTIF clipsMouseEventLeftUp, "clipsMouseEventLeftUp");
  DefineFunction("mouse-event-middle-up", 'l', PTIF clipsMouseEventMiddleUp, "clipsMouseEventMiddleUp");
  DefineFunction("mouse-event-right-up", 'l', PTIF clipsMouseEventRightUp, "clipsMouseEventRightUp");
  DefineFunction("mouse-event-dragging", 'l', PTIF clipsMouseEventDragging, "clipsMouseEventDragging");
  DefineFunction("mouse-event-position-x", 'd', PTIF clipsMouseEventPositionX, "clipsMouseEventPositionX");
  DefineFunction("mouse-event-position-y", 'd', PTIF clipsMouseEventPositionY, "clipsMouseEventPositionY");

  DefineFunction("font-create", 'l', PTIF clipsFontCreate, "clipsFontCreate");
  DefineFunction("font-delete", 'l', PTIF clipsFontDelete, "clipsFontDelete");

  DefineFunction("frame-create", 'l', PTIF clipsFrameCreate, "clipsFrameCreate");
  DefineFunction("frame-create-status-line", 'l', PTIF clipsFrameCreateStatusLine, "clipsFrameCreateStatusLine");
  DefineFunction("frame-iconize", 'l', PTIF clipsFrameIconize, "clipsFrameIconize");
  DefineFunction("frame-set-icon", 'l', PTIF clipsFrameSetIcon, "clipsFrameSetIcon");
  DefineFunction("frame-set-menu-bar", 'l', PTIF clipsFrameSetMenuBar, "clipsFrameSetMenuBar");
  DefineFunction("frame-set-status-text", 'l', PTIF clipsFrameSetStatusText, "clipsFrameSetStatusText");

  DefineFunction("icon-create", 'l', PTIF clipsIconCreate, "clipsIconCreate");
  DefineFunction("icon-delete", 'l', PTIF clipsIconDelete, "clipsIconDelete");

  DefineFunction("list-box-create", 'l', PTIF clipsListBoxCreate, "clipsListBoxCreate");
  DefineFunction("list-box-append", 'l', PTIF clipsListBoxAppend, "clipsListBoxAppend");
  DefineFunction("list-box-find-string", 'l', PTIF clipsListBoxFindString, "clipsListBoxFindString");
  DefineFunction("list-box-clear", 'l', PTIF clipsListBoxClear, "clipsListBoxClear");
  DefineFunction("list-box-set-selection", 'l', PTIF clipsListBoxSetSelection, "clipsListBoxSetSelection");
  DefineFunction("list-box-get-string-selection", 's', PTIF clipsListBoxGetStringSelection, "clipsListBoxGetStringSelection");
  DefineFunction("list-box-get-selection", 'l', PTIF clipsListBoxGetSelection, "clipsListBoxGetSelection");
  DefineFunction("list-box-set-string-selection", 'l', PTIF clipsListBoxSetStringSelection, "clipsListBoxSetStringSelection");
  DefineFunction("list-box-number", 'l', PTIF clipsListBoxNumber, "clipsListBoxNumber");
  DefineFunction("list-box-delete", 'l', PTIF clipsListBoxDelete, "clipsListBoxDelete");
  DefineFunction("list-box-get-string", 's', PTIF clipsListBoxGetString, "clipsListBoxGetString");
  DefineFunction("list-box-get-first-selection", 'l', PTIF clipsListBoxGetFirstSelection, "clipsListBoxGetFirstSelection");
  DefineFunction("list-box-get-next-selection", 'l', PTIF clipsListBoxGetNextSelection, "clipsListBoxGetNextSelection");

  DefineFunction("menu-create", 'l', PTIF clipsMenuCreate, "clipsMenuCreate");
  DefineFunction("menu-append", 'l', PTIF clipsMenuAppend, "clipsMenuAppend");
  DefineFunction("menu-append-separator", 'l', PTIF clipsMenuAppendSeparator, "clipsMenuAppendSeparator");
  DefineFunction("menu-check", 'l', PTIF clipsMenuCheck, "clipsMenuCheck");
  DefineFunction("menu-enable", 'l', PTIF clipsMenuEnable, "clipsMenuEnable");

  DefineFunction("menu-bar-create", 'l', PTIF clipsMenuBarCreate, "clipsMenuBarCreate");
  DefineFunction("menu-bar-append", 'l', PTIF clipsMenuBarAppend, "clipsMenuBarAppend");
  DefineFunction("menu-bar-check", 'l', PTIF clipsMenuBarCheck, "clipsMenuBarCheck");
  DefineFunction("menu-bar-enable", 'l', PTIF clipsMenuBarEnable, "clipsMenuBarEnable");

  DefineFunction("message-create", 'l', PTIF clipsMessageCreate, "clipsMessageCreate");

  DefineFunction("multi-text-create", 'l', PTIF clipsMultiTextCreate, "clipsMultiTextCreate");
  DefineFunction("multi-text-set-value", 'l', PTIF clipsMultiTextSetValue, "clipsMultiTextSetValue");
  DefineFunction("multi-text-get-value", 's', PTIF clipsMultiTextGetValue, "clipsMultiTextGetValue");

  DefineFunction("panel-create", 'l', PTIF clipsPanelCreate, "clipsPanelCreate");
  DefineFunction("panel-set-label-position", 'l', PTIF clipsPanelSetLabelPosition, "clipsPanelSetLabelPosition");
  DefineFunction("panel-new-line", 'l', PTIF clipsPanelNewLine, "clipsPanelNewLine");

  DefineFunction("panel-item-set-default", 'l', PTIF clipsPanelItemSetDefault, "clipsPanelItemSetDefault");
  DefineFunction("panel-item-set-label", 'l', PTIF clipsPanelItemSetLabel, "clipsPanelItemSetLabel");
  DefineFunction("panel-item-get-label", 's', PTIF clipsPanelItemGetLabel, "clipsPanelItemGetLabel");

  DefineFunction("pen-create", 'l', PTIF clipsPenCreate, "clipsPenCreate");
  DefineFunction("pen-delete", 'l', PTIF clipsPenDelete, "clipsPenDelete");

  DefineFunction("return-result", 'v', PTIF clipsReturnResult, "clipsReturnResult");

  DefineFunction("slider-create", 'l', PTIF clipsSliderCreate, "clipsSliderCreate");
  DefineFunction("slider-set-value", 'l', PTIF clipsSliderSetValue, "clipsSliderSetValue");
  DefineFunction("slider-get-value", 'l', PTIF clipsSliderGetValue, "clipsSliderGetValue");

  DefineFunction("string-sort", 'm', PTIF clipsStringSort, "clipsStringSort");

  DefineFunction("text-create", 'l', PTIF clipsTextCreate, "clipsTextCreate");
  DefineFunction("text-set-value", 'l', PTIF clipsTextSetValue, "clipsTextSetValue");
  DefineFunction("text-get-value", 's', PTIF clipsTextGetValue, "clipsTextGetValue");

  DefineFunction("text-window-create", 'l', PTIF clipsTextWindowCreate, "clipsTextWindowCreate");
  DefineFunction("text-window-load-file", 'l', PTIF clipsTextWindowLoadFile, "clipsTextWindowLoadFile");
  DefineFunction("text-window-save-file", 'l', PTIF clipsTextWindowSaveFile, "clipsTextWindowSaveFile");
  DefineFunction("text-window-modified", 'l', PTIF clipsTextWindowModified, "clipsTextWindowModified");
  DefineFunction("text-window-discard-edits", 'v', PTIF clipsTextWindowDiscardEdits, "clipsTextWindowDiscardEdits");
  DefineFunction("text-window-write", 'v', PTIF clipsTextWindowWrite, "clipsTextWindowWrite");

  DefineFunction("window-add-callback", 'l', PTIF clipsWindowAddCallback, "clipsWindowAddCallback");
  DefineFunction("window-delete", 'l', PTIF clipsWindowDelete, "clipsWindowDelete");
  DefineFunction("window-get-parent", 'l', PTIF clipsWindowGetParent, "clipsWindowGetParent");
  DefineFunction("window-show", 'l', PTIF clipsWindowShow, "clipsWindowShow");

  DefineFunction("window-get-x", 'l', PTIF clipsWindowGetX, "clipsWindowGetX");
  DefineFunction("window-get-y", 'l', PTIF clipsWindowGetY, "clipsWindowGetY");
  DefineFunction("window-get-width", 'l', PTIF clipsWindowGetWidth, "clipsWindowGetWidth");
  DefineFunction("window-get-height", 'l', PTIF clipsWindowGetHeight, "clipsWindowGetHeight");
  DefineFunction("window-get-client-width", 'l', PTIF clipsWindowGetClientWidth, "clipsWindowGetClientWidth");
  DefineFunction("window-get-client-height", 'l', PTIF clipsWindowGetClientHeight, "clipsWindowGetClientHeight");
  DefineFunction("window-set-size", 'l', PTIF clipsWindowSetSize, "clipsWindowSetSize");
  DefineFunction("window-set-client-size", 'l', PTIF clipsWindowSetClientSize, "clipsWindowSetClientSize");
  DefineFunction("window-set-focus", 'l', PTIF clipsWindowSetFocus, "clipsWindowSetFocus");
  DefineFunction("window-centre", 'l', PTIF clipsWindowCentre, "clipsWindowCentre");
  DefineFunction("window-fit", 'l', PTIF clipsWindowFit, "clipsWindowFit");

/*
  DefineFunction("help-load-file", 'l', PTIF clipsHelpLoadFile, "clipsHelpLoadFile");
  DefineFunction("help-display-contents", 'l', PTIF clipsHelpDisplayContents, "clipsHelpDisplayContents");
  DefineFunction("help-display-section", 'l', PTIF clipsHelpDisplaySection, "clipsHelpDisplaySection");
  DefineFunction("help-display-block", 'l', PTIF clipsHelpDisplayBlock, "clipsHelpDisplayBlock");
  DefineFunction("help-keyword-search", 'l', PTIF clipsHelpKeywordSearch, "clipsHelpKeywordSearch");
*/
}

ClipsErrorClass& ClipsErrorClass::operator << (char *s)
{
  ClipsErrorFunction(s);
  return *this;
}

ClipsErrorClass& ClipsErrorClass::operator << (char c)
{
  char buf[2];
  buf[0] = c;
  buf[1] = 0;

  ClipsErrorFunction(buf);
  return *this;
}

ClipsErrorClass& ClipsErrorClass::operator << (int i)
{
  char buf[20];
  sprintf(buf, "%d", i);

  ClipsErrorFunction(buf);
  return *this;
}

ClipsErrorClass& ClipsErrorClass::operator << (long i)
{
  char buf[20];
  sprintf(buf, "%ld", i);

  ClipsErrorFunction(buf);
  return *this;
}

ClipsErrorClass& ClipsErrorClass::operator << (double d)
{
  char buf[20];
  sprintf(buf, "%.2f", d);

  ClipsErrorFunction(buf);
  return *this;
}

ClipsErrorClass& ClipsErrorClass::operator << (float f)
{
  char buf[20];
  sprintf(buf, "%.2f", f);

  ClipsErrorFunction(buf);
  return *this;
}

void wxInitClips(void)
{
  InitializeCLIPS();
}

/*
 * Storage/retrieval functions.
 */

void wxStoreThingInTable(long item_id, wxObject *object, WXTYPE type)
{
#ifdef wx_x
  char buf[100];
  wxTypedObject *old_obj = (wxTypedObject *)wxClipsObjects.Get(item_id);
  if (old_obj)
  {
    sprintf(buf, "Warning: hash table clash,  item = %ld, old type = %d, new type = %d", item_id, old_obj->type, type);
    return;
  }
#endif
  wxTypedObject *obj = new wxTypedObject;
  obj->object = object;
  obj->type = type;
  wxClipsObjects.Put(item_id, obj);
}

void wxDeleteThingInTable(long item_id)
{
  wxTypedObject *obj = (wxTypedObject *)wxClipsObjects.Delete(item_id);
  delete obj;
}

wxObject *wxGetThingFromTable(long item_id, WXTYPE *type)
{
  wxTypedObject *obj = (wxTypedObject *)wxClipsObjects.Get(item_id);
  if (!obj)
    return NULL;

  if (type)
    *type = obj->type;
  return obj->object;
}

wxObject *wxGetTypedObject(long objectId,
                           WXTYPE expectedType)
{
  WXTYPE type;
  wxObject *obj = wxGetThingFromTable(objectId, &type);
  if (!obj)
    return NULL;
  
  if (!wxSubType(type, expectedType))
  {
    char buf[300];
    sprintf(buf, "Type error: %s found where %s expected.\n",
            wxGetTypeName(type), wxGetTypeName(expectedType));
    ClipsError << buf;
    return NULL;
  }
  return obj;
}
